import os

class WorkInTestDir:
    """
    Simple class that enables working in test directory
    """
    def __enter__(self):
        self.current_work_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.abspath(__file__)))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.current_work_dir)


import unittest
import unittest.mock

from mcstasscript.helper.mcstas_objects import provide_parameter
from mcstasscript.helper.mcstas_objects import write_parameter

class Test_ParameterVariable(unittest.TestCase):
    """
    Tests the ParameterVariable class that holds an input parameter
    for the instrument.

    """

    def test_ParameterVariable_init_basic(self):
        """
        Smallest possible initialization
        """

        par = provide_parameter("test")
        self.assertEqual(par.name, "test")

    def test_ParameterVariable_init_basic_type(self):
        """
        Initialization with a type
        """

        par = provide_parameter("double", "test")

        self.assertEqual(par.name, "test")
        self.assertEqual(par.type, "double")

    def test_ParameterVariable_init_basic_type_value(self):
        """
        Initialization with type and value
        """

        par = provide_parameter("double", "test", value=518)

        self.assertEqual(par.name, "test")
        self.assertEqual(par.type, "double")
        self.assertEqual(par.value, 518)

    def test_ParameterVariable_init_basic_type_value_comment(self):
        """
        Initialization with type, value and comment
        """

        par = provide_parameter("double", "test", value=518,
                                comment="test comment /")

        self.assertEqual(par.name, "test")
        self.assertEqual(par.type, "double")
        self.assertEqual(par.value, 518)
        self.assertEqual(par.comment, "test comment /")

    def test_ParameterVariable_init_basic_value_comment(self):
        """
        Initialization with value and comment
        """

        par = provide_parameter("test", value=518,
                                comment="test comment /")

        self.assertEqual(par.name, "test")
        self.assertEqual(par.type, "")
        self.assertEqual(par.value, 518)
        self.assertEqual(par.comment, "test comment /")

    def test_ParameterVariable_init_options_initialize(self):
        """
        Initialization with value and comment
        """

        par = provide_parameter("test", value=2,
                                options=[1, 2, 3.1])

        self.assertEqual(par.name, "test")
        self.assertEqual(par.type, "")
        self.assertEqual(par.value, 2)
        self.assertEqual(par.get_options(), [1, 2, 3.1])

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_ParameterVariable_write_basic(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in an
        instrument parameter section. The write file operation is
        mocked and check using a patch. Here a simple parameter is
        used.
        """

        par = provide_parameter("double", "test")
        with mock_f('test.txt', 'w') as m_fo:
            write_parameter(m_fo, parameter=par, stop_character="")

        expected_writes = [unittest.mock.call("double test"),
                           unittest.mock.call(""),
                           unittest.mock.call(""),
                           unittest.mock.call("\n")]

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        handle.write.assert_has_calls(expected_writes, any_order=False)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_ParameterVariable_write_complex_float(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in an
        instrument parameter section. The write file operation is
        mocked and check using a patch. Here a parameter with a value
        is used. (float value)
        """

        par = provide_parameter("double", "test", value=5.4,
                                comment="test comment")

        with mock_f('test.txt', 'w') as m_fo:
            write_parameter(m_fo, parameter=par, stop_character=")")

        expected_writes = [unittest.mock.call("double test"),
                           unittest.mock.call(" = 5.4"),
                           unittest.mock.call(")"),
                           unittest.mock.call("// test comment"),
                           unittest.mock.call("\n")]

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        handle.write.assert_has_calls(expected_writes, any_order=False)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_ParameterVariable_write_complex_int(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in an
        instrument parameter section. The write file operation is
        mocked and check using a patch. Here a parameter with a value
        is used. (integer value)
        """

        par = provide_parameter("double", "test", value=5,
                                comment="test comment")

        with mock_f('test.txt', 'w') as m_fo:
            write_parameter(m_fo, parameter=par, stop_character=")")

        expected_writes = [unittest.mock.call("double test"),
                           unittest.mock.call(" = 5"),
                           unittest.mock.call(")"),
                           unittest.mock.call("// test comment"),
                           unittest.mock.call("\n")]

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        handle.write.assert_has_calls(expected_writes, any_order=False)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_ParameterVariable_write_complex_string(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in an
        instrument parameter section. The write file operation is
        mocked and check using a patch. Here a parameter with a value
        is used. (string value)
        """

        par = provide_parameter("double", "test", value="\"Al\"",
                                comment="test comment")

        with mock_f('test.txt', 'w') as m_fo:
            write_parameter(m_fo, parameter=par, stop_character=",")

        expected_writes = [unittest.mock.call("double test"),
                           unittest.mock.call(" = \"Al\""),
                           unittest.mock.call(","),
                           unittest.mock.call("// test comment"),
                           unittest.mock.call("\n")]

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        handle.write.assert_has_calls(expected_writes, any_order=False)


if __name__ == '__main__':
    unittest.main()



import unittest
import unittest.mock

from mcstasscript.helper.mcstas_objects import DeclareVariable


class Test_DeclareVariable(unittest.TestCase):
    """
    Tests the DeclareVariable class that holds a declared variable
    that will be written to the McStas declare section.

    """

    def test_DeclareVariable_init_basic_type(self):
        """
        Initialization with a type
        """

        var = DeclareVariable("double", "test")

        self.assertEqual(var.name, "test")
        self.assertEqual(var.type, "double")

    def test_DeclareVariable_init_basic_type_value(self):
        """
        Initialization with type and value
        """

        var = DeclareVariable("double", "test", value=518)

        self.assertEqual(var.name, "test")
        self.assertEqual(var.type, "double")
        self.assertEqual(var.value, 518)

    def test_DeclareVariable_init_basic_type_vector(self):
        """
        Initialization with type and value
        """

        var = DeclareVariable("double", "test",
                              array=6, value=[1, 2.2, 3, 3.3, 4, 4.4])

        self.assertEqual(var.name, "test")
        self.assertEqual(var.type, "double")  # space for easier writing
        self.assertEqual(var.vector, 6)
        self.assertEqual(var.value, [1, 2.2, 3, 3.3, 4, 4.4])

    def test_DeclareVariable_init_basic_type_value_comment(self):
        """
        Initialization with type, value and comment
        """

        var = DeclareVariable("double", "test",
                              value=518, comment="test comment /")

        self.assertEqual(var.name, "test")
        self.assertEqual(var.type, "double")  # Space for easier writing
        self.assertEqual(var.value, 518)
        self.assertEqual(var.comment, " // test comment /")

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_DeclareVariable_write_basic(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in a
        instrument declare section. The write file operation is
        mocked and check using a patch. Here a simple declare is
        used.
        """

        var = DeclareVariable("double", "test")
        with mock_f('test.txt', 'w') as m_fo:
            var.write_line(m_fo)

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        handle.write.assert_called_once_with("double test;")

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_DeclareVariable_write_complex_float(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in an
        instrument declare section. The write file operation is
        mocked and check using a patch. Here a declare with a value
        is used. (float value)
        """

        var = DeclareVariable("double",
                              "test",
                              value=5.4,
                              comment="test comment")

        with mock_f('test.txt', 'w') as m_fo:
            var.write_line(m_fo)

        expected_write = "double test = 5.4; // test comment"

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        handle.write.assert_called_once_with(expected_write)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_DeclareVariable_write_complex_int(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in an
        instrument declare section. The write file operation is
        mocked and check using a patch. Here a declare with a value
        is used. (integer value)
        """

        var = DeclareVariable("double",
                              "test",
                              value=5,
                              comment="test comment")

        with mock_f('test.txt', 'w') as m_fo:
            var.write_line(m_fo)

        expected_write = "double test = 5; // test comment"

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        handle.write.assert_called_once_with(expected_write)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_DeclareVariable_write_simple_array(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in an
        instrument declare section. The write file operation is
        mocked and check using a patch. Here an array is declared.
        """

        var = DeclareVariable("double",
                              "test",
                              array=29,
                              comment="test comment")

        with mock_f('test.txt', 'w') as m_fo:
            var.write_line(m_fo)

        expected_write = "double test[29]; // test comment"

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        handle.write.assert_called_once_with(expected_write)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_DeclareVariable_write_complex_array(self, mock_f):
        """
        Testing that write to file is correct. Here a line is in an
        instrument declare section. The write file operation is
        mocked and check using a patch. Here an array is declared and
        populated with the selected values.
        """

        var = DeclareVariable("double",
                              "test",
                              array=3,
                              value=[5, 4, 3.1],
                              comment="test comment")

        with mock_f('test.txt', 'w') as m_fo:
            var.write_line(m_fo)

        expected_writes = [unittest.mock.call("double test[3] = {5,4,3.1}; // test comment")]

        mock_f.assert_called_with('test.txt', 'w')
        handle = mock_f()
        for c, w in zip(handle.write.call_args_list, expected_writes):
            assert(c == w)
        handle.write.assert_has_calls(expected_writes, any_order=False)


if __name__ == '__main__':
    unittest.main()



import io
import os
import unittest.mock

from mcstasscript.interface import instr


def setup_simple_instrument():
    Instr = instr.McStas_instr("integration_test_simple")

    source = Instr.add_component("source", "Source_div")

    source.xwidth = 0.03
    source.yheight = 0.01
    source.focus_aw = 0.01
    source.focus_ah = 0.01
    source.E0 = 81.81
    source.dE = 1.0
    source.flux = 1E10

    PSD = Instr.add_component("PSD_1D", "PSDlin_monitor")

    PSD.set_AT([0, 0, 1], RELATIVE="source")
    PSD.xwidth = 0.1
    if Instr.mccode_version > 2:
        PSD.nbins = 100
    else:
        PSD.nx = 100
    PSD.yheight = 0.03
    PSD.filename = "\"PSD.dat\""
    PSD.restore_neutron = 1

    return Instr


def setup_simple_instrument_input_path():
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    input_path = os.path.join(THIS_DIR, "test_input_folder")

    Instr = instr.McStas_instr("integration_test_simple_input",
                               input_path=input_path)

    source = Instr.add_component("source", "Source_div")

    source.xwidth = 0.03
    source.yheight = 0.01
    source.focus_aw = 0.01
    source.focus_ah = 0.01
    source.E0 = 81.81
    source.dE = 1.0
    source.flux = 1E10

    if Instr.mccode_version > 2:
        PSD = Instr.add_component("PSD_1D", "PSDlin_monitor_3")
    else:
        PSD = Instr.add_component("PSD_1D", "PSDlin_monitor")

    PSD.set_AT([0, 0, 1], RELATIVE="source")
    PSD.xwidth = 0.1
    if Instr.mccode_version > 2:
        PSD.nbins = 100
    else:
        PSD.nx = 100
    PSD.yheight = 0.03
    PSD.filename = "\"PSD.dat\""
    PSD.restore_neutron = 1

    return Instr


def setup_simple_slit_instrument():
    Instr = instr.McStas_instr("integration_test_simple")

    source = Instr.add_component("source", "Source_div")
    source.xwidth = 0.1
    source.yheight = 0.01
    source.focus_aw = 0.01
    source.focus_ah = 0.01
    source.E0 = 81.81
    source.dE = 1.0
    source.flux = 1E10

    Instr.add_parameter("slit_offset", value=0)

    Slit = Instr.add_component("slit", "Slit")
    Slit.set_AT(["slit_offset", 0, 0.5], RELATIVE="source")
    Slit.xwidth = 0.01
    Slit.yheight = 0.03

    PSD = Instr.add_component("PSD_1D", "PSDlin_monitor")
    PSD.set_AT([0, 0, 1], RELATIVE="source")
    PSD.xwidth = 0.1
    if Instr.mccode_version > 2:
        PSD.nbins = 100
    else:
        PSD.nx = 100
    PSD.yheight = 0.03
    PSD.filename = "\"PSD.dat\""
    PSD.restore_neutron = 1

    return Instr


class TestSimpleInstrument(unittest.TestCase):
    """
    Integration test of a full instrument with McStas simulation
    performed by the system. The configuration file needs to be set up
    correctly in order for these tests to succeed.
    """

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_simple_instrument(self, mock_stdout):
        """
        Test that an instrument can run and that the results match
        expectations. Here beam in small area in the middle of the
        detector.
        """
        CURRENT_DIR = os.getcwd()
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        os.chdir(THIS_DIR)

        Instr = setup_simple_instrument()

        data = Instr.run_full_instrument(foldername="integration_test_simple",
                                         ncount=1E6, mpi=1,
                                         increment_folder_name=True)

        os.chdir(CURRENT_DIR)

        intensity_data = data[0].Intensity
        # beam should be on pixel 35 to 65

        sum_outside_beam = (sum(intensity_data[0:34])
                            + sum(intensity_data[66:99]))
        sum_inside_beam = sum(intensity_data[35:65])

        self.assertTrue(1000*sum_outside_beam < sum_inside_beam)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_simple_instrument_input(self, mock_stdout):
        """
        Test that an instrument can run and that the results match
        expectations. Here beam in small area in the middle of the
        detector.
        """
        CURRENT_DIR = os.getcwd()
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        os.chdir(THIS_DIR)

        Instr = setup_simple_instrument_input_path()

        foldername = "integration_test_simple_input"
        data = Instr.run_full_instrument(foldername=foldername,
                                         ncount=1E6, mpi=1,
                                         increment_folder_name=True)

        os.chdir(CURRENT_DIR)

        intensity_data = data[0].Intensity
        # beam should be on pixel 35 to 65

        sum_outside_beam = (sum(intensity_data[0:34])
                            + sum(intensity_data[66:99]))
        sum_inside_beam = sum(intensity_data[35:65])

        self.assertTrue(1000*sum_outside_beam < sum_inside_beam)

        # Check component from input_folder read
        self.assertEqual(data[0].metadata.xlabel, "Test")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_simple_instrument_mpi(self, mock_stdout):
        """
        Test that an instrument can run and that the results matches
        expectations. Here beam in small area in the middle of the
        detector. Running with mpi, 2 cores.
        """
        CURRENT_DIR = os.getcwd()
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        os.chdir(THIS_DIR)

        Instr = setup_simple_instrument()

        data = Instr.run_full_instrument(foldername="integration_test_mpi",
                                         ncount=1E6, mpi=2,
                                         increment_folder_name=True)

        os.chdir(CURRENT_DIR)

        intensity_data = data[0].Intensity
        # beam should be on pixel 35 to 65

        sum_outside_beam = (sum(intensity_data[0:34])
                            + sum(intensity_data[66:99]))
        sum_inside_beam = sum(intensity_data[35:65])

        self.assertTrue(1000*sum_outside_beam < sum_inside_beam)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_slit_instrument(self, mock_stdout):
        """
        Test parameters can be controlled through McStasScript.  Here
        a slit can be moved, but the default value of 0 should be
        used.
        """
        CURRENT_DIR = os.getcwd()
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        os.chdir(THIS_DIR)

        Instr = setup_simple_slit_instrument()

        foldername = "integration_test_slit"
        data = Instr.run_full_instrument(foldername=foldername,
                                         ncount=2E6, mpi=2,
                                         increment_folder_name=True)

        os.chdir(CURRENT_DIR)

        intensity_data = data[0].Intensity
        # beam should be on pixel 45 to 55

        sum_outside_beam = (sum(intensity_data[0:44])
                            + sum(intensity_data[56:99]))
        sum_inside_beam = sum(intensity_data[45:55])
        self.assertTrue(1000*sum_outside_beam < sum_inside_beam)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_slit_moved_instrument(self, mock_stdout):
        """
        Test parameters can be controlled through McStasScript.  Here
        a slit is moved to one side and the result is verified.
        """
        CURRENT_DIR = os.getcwd()
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        os.chdir(THIS_DIR)

        Instr = setup_simple_slit_instrument()

        data = Instr.run_full_instrument(foldername="integration_test_slit",
                                         ncount=2E6, mpi=2,
                                         increment_folder_name=True,
                                         parameters={"slit_offset": 0.03})

        os.chdir(CURRENT_DIR)

        intensity_data = data[0].Intensity
        # beam should be on pixel 75 to 85

        sum_outside_beam = (sum(intensity_data[0:74])
                            + sum(intensity_data[86:99]))
        sum_inside_beam = sum(intensity_data[75:85])

        self.assertTrue(1000*sum_outside_beam < sum_inside_beam)


if __name__ == '__main__':
    unittest.main()



from mcstasscript.instrument_diagram.connections import ConnectionList
from mcstasscript.instrument_diagram.arrow import Arrow


def generate_AT_arrows(components, component_box_dict, box_names, color=None):
    """
    Generate Arrow objects related to the AT relationship of components
    """
    connections = ConnectionList()
    for component in components:
        origin = component_box_dict[component.name]
        if component.AT_reference is None:
            target = component_box_dict["ABSOLUTE"]
        elif component.AT_reference == "PREVIOUS":
            origin_index = box_names.index(component.name)
            target_index = origin_index - 1
            target = component_box_dict[box_names[target_index]]
        else:
            target = component_box_dict[component.AT_reference]

        connections.add(origin, target)

    connections.distribute_lane_numbers(box_names=box_names)

    arrows = []
    for connection in connections.get_connections():
        origin = connection.origin
        target = connection.target
        lane = connection.lane_number

        arrow = Arrow(origin, target, lane=lane, kind="AT")

        arrow.set_box_offset_origin(0.24)
        arrow.set_box_offset_target(-0.2)
        if color is None:
            arrow.color = "blue"
        else:
            arrow.color = color
        # arrow.set_linestyle("--")

        arrows.append(arrow)

    return arrows



class IndexConnection:
    def __init__(self, start_index, end_index):
        self.start_index = start_index
        self.end_index = end_index

        if start_index <= end_index:
            self.interval_start = start_index
            self.interval_end = end_index
        else:
            self.interval_start = end_index
            self.interval_end = start_index

    def compatible_with(self, new):
        """
        Check if the new interval can coexist with the existing
        """
        if new.interval_end > self.interval_start and new.interval_start < self.interval_end:
            return False

        return True


class Lane:
    def __init__(self):
        self.connections = []

    def add_connection(self, start_index, end_index):
        """
        If possible adds connection and returns True, otherwise returns False
        """
        new_connection = IndexConnection(start_index, end_index)

        # A line can skip on either start or end, not both
        skipped_end = False
        skipped_start = False

        # Check if there is room for this lane
        for connection in self.connections:
            # Check for number of reasons for connection being allowed
            if connection.start_index == new_connection.start_index and not skipped_end:
                # Allow connections to collide when the start index matches
                skipped_start = True
                continue

            if connection.end_index == new_connection.end_index and not skipped_start:
                # Allow connections to collide when the end index matches
                skipped_end = True
                continue

            if connection.compatible_with(new_connection):
                # Allow if there are no index overlap in lane
                continue

            # If connection incompatible, return false
            return False

        # No problems, this connection can be included in this lane
        self.connections.append(new_connection)
        return True


class Connection:
    def __init__(self, origin, target, info=None):
        """
        Describes a connection between origin and target with lane number

        Can contain info as well which can be used for example to mark
        if the connection is part of a certain group.
        """
        self.origin = origin
        self.target = target
        self.lane_number = None
        self.info = info

    def set_lane_number(self, value):
        self.lane_number = value


class ConnectionList:
    def __init__(self):
        """
        List of connections with utility functions

        Can distribute the connections over a number of lanes to ensure all
        connections can be seen without crossing that provide ambitious
        interpretations.
        """
        self.connections = []

    def add(self, origin, target, info=None):
        self.connections.append(Connection(origin, target, info=info))

    def get_connections(self):
        return self.connections

    def get_origins(self):
        return [x.origin for x in self.connections]

    def get_targets(self):
        return [x.target for x in self.connections]

    def get_pairs(self):
        return zip(self.get_origins(), self.get_targets())

    def get_targets_for_origin(self, given_origin):

        return_targets = []
        for origin, target in self.get_pairs():
            if origin == given_origin:
                return_targets.append(target)

        return return_targets

    def distribute_lane_numbers(self, box_names):
        lanes = []

        for connection in self.connections:
            start_index = box_names.index(connection.origin.name)
            end_index = box_names.index(connection.target.name)

            proposed_lane = 0
            while True:
                if proposed_lane >= len(lanes):
                    lanes.append(Lane())

                if lanes[proposed_lane].add_connection(start_index=start_index, end_index=end_index):
                    connection.set_lane_number(proposed_lane + 1)
                    break

                proposed_lane += 1




import os
import unittest


import numpy as np

from mcstasscript.interface.functions import name_search
from mcstasscript.interface.functions import name_plot_options
from mcstasscript.interface.functions import load_data
from mcstasscript.interface.functions import load_metadata
from mcstasscript.interface.functions import load_monitor
from mcstasscript.data.data import McStasDataBinned
from mcstasscript.data.data import McStasMetaData


def set_dummy_MetaDataBinned_1d(name):
    """
    Sets up a dummy MetaData object for a 1d dataset

    Parameters
    ----------

    name : str
        base for filename, .dat will be appended
    """

    meta_data = McStasMetaData()
    meta_data.component_name = name
    meta_data.dimension = 50
    meta_data.filename = name + ".dat"

    return meta_data


def set_dummy_McStasDataBinned_1d(name):
    """
    Sets up a dummy McStasData object for a 1d dataset

    Parameters
    ----------

    name : str
        base for filename, .dat will be appended
    """
    meta_data = set_dummy_MetaDataBinned_1d(name)

    intensity = np.arange(20)
    error = 0.5 * np.arange(20)
    ncount = 2 * np.arange(20)
    axis = np.arange(20)*5.0

    return McStasDataBinned(meta_data, intensity, error, ncount, xaxis=axis)


def set_dummy_MetaDataBinned_2d(name):
    """
    Sets up a dummy MetaData object for a 2d dataset

    Parameters
    ----------

    name : str
        base for filename, .dat will be appended
    """
    meta_data = McStasMetaData()
    meta_data.component_name = name
    meta_data.dimension = [50, 100]
    meta_data.filename = name + ".dat"

    return meta_data


def set_dummy_McStasDataBinned_2d(name):
    """
    Sets up a dummy McStasData object for a 2d dataset

    Parameters
    ----------

    name : str
        base for filename, .dat will be appended
    """

    meta_data = set_dummy_MetaDataBinned_2d(name)

    intensity = np.arange(20).reshape(4, 5)
    error = 0.5 * np.arange(20).reshape(4, 5)
    ncount = 2 * np.arange(20).reshape(4, 5)

    return McStasDataBinned(meta_data, intensity, error, ncount)


def setup_McStasData_array():
    """
    Sets up an list of McStasData objects, similar to simulation output
    """

    data_list = []

    data_list.append(set_dummy_McStasDataBinned_1d("A_1d_thing"))
    data_list.append(set_dummy_McStasDataBinned_2d("A_2d_thing"))
    data_list.append(set_dummy_McStasDataBinned_1d("Another_1d_thing"))
    data_list.append(set_dummy_McStasDataBinned_2d("Another_2d_thing"))
    data_list.append(set_dummy_McStasDataBinned_2d("A_third_2d_thing"))

    hero_object = set_dummy_McStasDataBinned_2d("Hero")
    hero_object.metadata.dimension = 123
    hero_object.plot_options.colormap = "very hot"

    data_list.append(hero_object)

    data_list.append(set_dummy_McStasDataBinned_2d("After_hero_2d"))
    data_list.append(set_dummy_McStasDataBinned_2d("Last_object_2d"))

    return data_list


def setup_McStasData_array_repeat():
    """
    Sets up an list of McStasData objects, similar to simulation output

    Have Hero twice in naming, testing search capability
    """

    data_list = []

    data_list.append(set_dummy_McStasDataBinned_1d("A_1d_thing"))
    data_list.append(set_dummy_McStasDataBinned_2d("A_2d_thing"))
    data_list.append(set_dummy_McStasDataBinned_1d("Another_1d_thing"))
    data_list.append(set_dummy_McStasDataBinned_2d("Another_2d_thing"))
    data_list.append(set_dummy_McStasDataBinned_2d("Hero"))

    hero_object = set_dummy_McStasDataBinned_2d("Big_Hero")
    hero_object.metadata.dimension = 123
    hero_object.plot_options.colormap = "very hot"

    data_list.append(hero_object)

    data_list.append(set_dummy_McStasDataBinned_2d("After_hero_2d"))
    data_list.append(set_dummy_McStasDataBinned_2d("Last_object_2d"))

    return data_list


class Test_name_search(unittest.TestCase):
    """
    Test the utility function called name_search which finds and
    returns a McStasData set with a given name from a list of
    McStasData objects.
    """

    def test_name_search_read(self):
        """
        Test that Hero object can be found and check the unique dimension

        Here the name is used
        """

        data_list = setup_McStasData_array()

        hero_object = name_search("Hero", data_list)

        self.assertEqual(hero_object.metadata.dimension, 123)

    def test_name_search_filename_read(self):
        """
        Test that Hero object can be found and check the unique dimension

        Here the name of the datafile is used
        """

        data_list = setup_McStasData_array()

        hero_object = name_search("Hero.dat", data_list)

        self.assertEqual(hero_object.metadata.dimension, 123)

    def test_name_search_read_repeat(self):
        """
        Test that Hero object can be found and check the unique dimension
        Here the used data set has two monitors with Hero in the name

        Here the name of the monitor is used
        """

        data_list = setup_McStasData_array_repeat()

        hero_object = name_search("Big_Hero", data_list)

        self.assertEqual(hero_object.metadata.dimension, 123)

    def test_name_search_read_duplicate(self):
        """
        Test simple case with duplicated name, search should return list
        """

        data_list = setup_McStasData_array_repeat()

        # Adds another dataset with a name already in the data_list
        hero_object = set_dummy_McStasDataBinned_2d("Big_Hero")
        hero_object.metadata.dimension = 321
        hero_object.plot_options.colormap = "very hot"

        data_list.append(hero_object)

        # Now two McStasData objects match the Big_Hero name
        results = name_search("Big_Hero", data_list)

        self.assertEqual(type(results), list)
        # Check two results are returned
        self.assertEqual(len(results), 2)

        # Check they have the correct dimensions
        self.assertEqual(results[0].metadata.dimension, 123)
        self.assertEqual(results[1].metadata.dimension, 321)

    def test_name_search_read_error(self):
        """
        Check an NameError is returned when no match is found
        """

        data_list = setup_McStasData_array()

        with self.assertRaises(NameError):
            name_search("Hero8", data_list)

    def test_name_search_type_error_not_list(self):
        """
        Check error is given even when data list is just single object
        """

        data_list = set_dummy_McStasDataBinned_2d("Last_object_2d")

        with self.assertRaises(RuntimeError):
            name_search("Hero", data_list)

    def test_name_search_type_error_not_McStasData(self):
        """
        Checks that an error is returned if the given dataset contains
        non McStasData objects
        """

        data_list = [1, 2, 3]

        with self.assertRaises(RuntimeError):
            name_search(1, data_list)


class Test_name_plot_options(unittest.TestCase):
    """
    Test the utility function called name_plot_options which sends
    keyword arguments to the set_plot_options method of the
    McStasData object in a given list that has the given name.

    """

    def test_name_plot_options_simple(self):
        """
        Check set_plot_options can modify given attribute
        """

        data_list = setup_McStasData_array()
        name_plot_options("Hero", data_list, colormap="Oranges")
        hero_object = name_search("Hero", data_list)
        self.assertEqual(hero_object.plot_options.colormap, "Oranges")

    def test_name_plot_options_duplicate(self):
        """
        Test case where several McStasData objects are modified since
        the internal name_search finds multiple matches
        """

        data_list = setup_McStasData_array()

        hero_object = set_dummy_McStasDataBinned_2d("Hero")
        hero_object.metadata.dimension = 321
        hero_object.plot_options.colormap = "absurdly hot"

        data_list.append(hero_object)

        name_plot_options("Hero", data_list, colormap="Blues")

        results = name_search("Hero", data_list)

        self.assertEqual(len(results), 2)
        self.assertEqual(results[0].plot_options.colormap, "Blues")
        self.assertEqual(results[1].plot_options.colormap, "Blues")


class Test_load_data(unittest.TestCase):
    """
    Testing the load data function which calls ManagedMcrun, which was
    tested elsewhere. Since the load data is tested elsewhere, this
    function has just a single test to check the interface.
    """
    def test_mcrun_load_data_PSD4PI(self):
        """
        Use test_data_set to test load_data for PSD_4PI
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        results = load_data("test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(len(results), 4)

        PSD_4PI = results[0]

        self.assertEqual(PSD_4PI.name, "PSD_4PI")
        self.assertEqual(PSD_4PI.metadata.dimension, [300, 300])
        self.assertEqual(PSD_4PI.metadata.limits, [-180, 180, -90, 90])
        self.assertEqual(PSD_4PI.metadata.xlabel, "Longitude [deg]")
        self.assertEqual(PSD_4PI.metadata.ylabel, "Latitude [deg]")
        self.assertEqual(PSD_4PI.metadata.title, "4PI PSD monitor")
        self.assertEqual(PSD_4PI.Ncount[4][1], 4)
        self.assertEqual(PSD_4PI.Intensity[4][1], 1.537334562E-10)
        self.assertEqual(PSD_4PI.Error[4][1], 1.139482296E-10)


class Test_load_metadata(unittest.TestCase):
    """
    Testing the load metadata function which calls ManagedMcrun, which
    was tested elsewhere. Since the load metadata is tested elsewhere,
    this function has just a single test to check the interface.
    """
    def test_mcrun_load_metadata_PSD4PI(self):
        """
        Use test_data_set to test load_metadata for PSD_4PI
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        metadata = load_metadata("test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(len(metadata), 4)

        PSD_4PI = metadata[0]
        self.assertEqual(PSD_4PI.dimension, [300, 300])
        self.assertEqual(PSD_4PI.limits, [-180, 180, -90, 90])
        self.assertEqual(PSD_4PI.xlabel, "Longitude [deg]")
        self.assertEqual(PSD_4PI.ylabel, "Latitude [deg]")
        self.assertEqual(PSD_4PI.title, "4PI PSD monitor")

class Test_load_monitor(unittest.TestCase):
    """
    Testing the load monitor function which calls ManagedMcrun, which
    was tested elsewhere. Since the load monitor is tested elsewhere, this
    function has just a single test to check the interface.
    """
    def test_mcrun_load_monitor_PSD4PI(self):
        """
        Use test_data_set to test load_monitor for PSD_4PI
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        metadata = load_metadata("test_data_set")
        PSD_4PI = metadata[0]
        monitor = load_monitor(PSD_4PI, "test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(monitor.name, "PSD_4PI")
        self.assertEqual(monitor.metadata.dimension, [300, 300])
        self.assertEqual(monitor.metadata.limits, [-180, 180, -90, 90])
        self.assertEqual(monitor.metadata.xlabel, "Longitude [deg]")
        self.assertEqual(monitor.metadata.ylabel, "Latitude [deg]")
        self.assertEqual(monitor.metadata.title, "4PI PSD monitor")
        self.assertEqual(monitor.Ncount[4][1], 4)
        self.assertEqual(monitor.Intensity[4][1], 1.537334562E-10)
        self.assertEqual(monitor.Error[4][1], 1.139482296E-10)


if __name__ == '__main__':
    unittest.main()



import sys
import os

class HiddenPrints:
    """
    Environment which suppress prints
    """
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def parameter_has_default(parameter):
    """
    Checks if ParameterVariable has a default value, returns bool

    Parameters
    ----------

    parameter: ParameterVariable
        The parameter to check for default value
    """
    if parameter.value is None:
        return False
    return True


def get_parameter_default(parameter):
    """
    Returns the default value of a parameter

    Parameters
    ----------

    parameter: ParameterVariable
        The parameter for which the default value is returned
    """
    if parameter.value is not None:
        if parameter.type == "string":
            return parameter.value
        elif parameter.type == "double" or parameter.type == "":
            return float(parameter.value)
        elif parameter.type == "int":
            return int(parameter.value)
        else:
            raise RuntimeError("Unknown parameter type '"
                               + parameter.type + "' of par named '"
                               + parameter.name + "'.")

    return None



import os
import unittest
import unittest.mock
import copy

from mcstasscript.instrument_diagram.arrow import Arrow
from mcstasscript.instrument_diagram.box import ComponentBox
from mcstasscript.instrument_diagram.connections import IndexConnection
from mcstasscript.instrument_diagram.connections import Lane
from mcstasscript.instrument_diagram.connections import ConnectionList
from mcstasscript.instrument_diagram.canvas import DiagramCanvas
from mcstasscript.tests.helpers_for_tests import WorkInTestDir
from mcstasscript.interface.instr import McStas_instr


def setup_instr_no_path():
    """
    Sets up a neutron instrument without a package_path
    """

    with WorkInTestDir() as handler:
        instrument = McStas_instr("test_instrument")

    return instrument

def setup_populated_instr():
    """
    Sets up a neutron instrument with some features used and three components
    """
    instr = setup_instr_no_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    A = instr.add_component("first_component", "test_for_reading")
    B = instr.add_component("second_component", "test_for_reading")
    C = instr.add_component("third_component", "test_for_reading")
    D = instr.add_component("fourth_component", "test_for_reading")

    B.set_AT(0, RELATIVE=A)
    C.set_AT(0, RELATIVE=A)
    D.set_AT(0, RELATIVE=B)

    return instr

def setup_boxes():

    instr = setup_populated_instr()

    comp1 = instr.get_component("first_component")
    comp2 = instr.get_component("third_component")

    return ComponentBox(comp1), ComponentBox(comp2)


class TestInstrumentDiagram(unittest.TestCase):
    """
    Instrument diagram is not critical for operation, tests just to ensure it runs
    """

    def test_Box_initialize_name(self):
        """
        Check that a box can be initialized with a string
        """
        name = "ABSOLUTE"
        box = ComponentBox(name)

        self.assertEqual(box.name, name)

    def test_Box_initialize_comp(self):
        """
        Check that a box can be initialized with a component object
        """
        instr = setup_populated_instr()
        comp = instr.get_last_component()

        box = ComponentBox(comp)

        self.assertEqual(box.name, "fourth_component")

    def test_Arrow_initialize(self):
        """
        Check that an arrow can be initialized with two Box instances
        """

        box1, box2 = setup_boxes()

        arrow = Arrow(origin=box1, target=box2, kind="KIND")

        self.assertEqual(arrow.origin, box1)
        self.assertEqual(arrow.target, box2)
        self.assertEqual(arrow.kind, "KIND")

        string_box = ComponentBox("string")
        arrow = Arrow(origin=box1, target=string_box, kind="KIND_2")

        self.assertEqual(arrow.origin, box1)
        self.assertEqual(arrow.target, string_box)
        self.assertEqual(arrow.kind, "KIND_2")

    def test_IndexConnection(self):
        """
        Ensure IndexConnections find overlap
        """

        A = IndexConnection(2, 8)
        B = IndexConnection(9, 10)
        C = IndexConnection(5, 7)

        self.assertTrue(A.compatible_with(B))
        self.assertTrue(B.compatible_with(C))
        self.assertFalse(A.compatible_with(C))

    def test_Lane(self):
        """
        Ensure a lane reports when a connection can be added and saves allowed
        """

        lane = Lane()

        self.assertTrue(lane.add_connection(2, 8))
        self.assertTrue(lane.add_connection(9, 12))
        self.assertTrue(lane.add_connection(7, 8))  # Same destination allowed
        self.assertFalse(lane.add_connection(3, 9))

        # ensure only allowed connections added
        self.assertEqual(len(lane.connections), 3)

    def test_connection(self):
        """
        Ensure connection lists can return included connections and set lanes
        """
        c_list = ConnectionList()

        box_list = []
        name_list = []
        for index in range(0,10):
            name = "box_" + str(index)
            name_list.append(name)
            box_list.append(ComponentBox(name))

        # Add some connections to list
        c_list.add(box_list[0], box_list[5])  # Lane 1 # Connection 0
        c_list.add(box_list[6], box_list[8])  # Lane 1 # Connection 1
        c_list.add(box_list[1], box_list[7])  # Lane 2 # Connection 2
        c_list.add(box_list[3], box_list[6])  # Lane 3 # Connection 3
        c_list.add(box_list[0], box_list[8])  # Lane 4 # Connection 4

        targets_for_0 = c_list.get_targets_for_origin(box_list[0])
        self.assertEqual(targets_for_0, [box_list[5], box_list[8]])

        c_list.distribute_lane_numbers(name_list)

        connections = c_list.get_connections()
        self.assertEqual(connections[0].lane_number, 1)
        self.assertEqual(connections[1].lane_number, 1)
        self.assertEqual(connections[2].lane_number, 2)
        self.assertEqual(connections[3].lane_number, 3)
        self.assertEqual(connections[4].lane_number, 4)

    def test_DiagramCanvas_initialize(self):
        """
        Ensure the DiagramCanvas can be initialized
        """
        string_box_0 = ComponentBox("string_0")
        string_box_1 = ComponentBox("string_1")
        string_box_2 = ComponentBox("string_2")
        arrow_0 = Arrow(origin=string_box_1, target=string_box_0)
        arrow_1 = Arrow(origin=string_box_2, target=string_box_0)
        arrow_0.set_lane(1)
        arrow_1.set_lane(1)

        left_side_arrows = [arrow_0, arrow_1]

        boxes = [string_box_0, string_box_1, string_box_2]

        arrow_0 = Arrow(origin=string_box_0, target=string_box_1)
        arrow_1 = Arrow(origin=string_box_1, target=string_box_0, kind="JUMP")
        arrow_2 = Arrow(origin=string_box_0, target=string_box_2, kind="Union")
        arrow_0.set_lane(1)
        arrow_1.set_lane(1)
        arrow_2.set_lane(2)
        right_side_arrows = [arrow_0, arrow_1, arrow_2]

        # Get component_categories
        with WorkInTestDir() as handler:
            instrument = McStas_instr("test_instrument")

        component_reader = instrument.component_reader
        component_categories = copy.deepcopy(component_reader.component_category)

        color_choices = {"AT": "blue", "ROTATED": "red", "JUMP": "black", "GROUP": [0.4, 0.4, 0.4], "Union": "green"}

        canvas = DiagramCanvas(left_side_arrows=left_side_arrows, component_boxes=boxes,
                               right_side_arrows=right_side_arrows,
                               component_categories=component_categories,
                               colors=color_choices)




import unittest
import os
import mcstasscript as ms
from mcstasscript.tests.helpers_for_tests import WorkInTestDir

from mcstasscript.tools.instrument_checker import has_component, has_parameter, all_parameters_set


def setup_instr_with_path():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder.
    """

    with WorkInTestDir() as handler:
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")
        instrument = ms.McStas_instr("test_instrument",
                                     package_path=dummy_path,
                                     executable_path=dummy_path)

    return instrument


def setup_populated_instr():
    """
    Sets up a neutron instrument with some features used and three components
    """
    instr = setup_instr_with_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_parameter("int", "n_pulses")

    instr.add_declare_var("double", "two_theta")

    instr.append_initialize("two_theta = 2.0*theta;")

    instr.add_component("first_component", "test_for_reading")
    instr.add_component("second_component", "test_for_reading")
    instr.add_component("third_component", "test_for_reading")

    return instr


class TestToolInstrumentChecker(unittest.TestCase):
    """
    Various test of instrument_checker functions
    """

    def test_simple_case_name(self):
        """
        check has_component of given name
        """
        instr = setup_populated_instr()
        self.assertTrue(has_component(instr, component_name="second_component"))

    def test_simple_case_name_not_found(self):
        """
        check has_component returns false when name is not there
        """
        instr = setup_populated_instr()
        self.assertFalse(has_component(instr, component_name="fourth_component"))

    def test_simple_case_type(self):
        """
        check has_component can find component of given type
        """
        instr = setup_populated_instr()
        self.assertTrue(has_component(instr, component_type="test_for_reading"))

    def test_simple_case_type_not_found(self):
        """
        check has_component returns false when component type is not found
        """
        instr = setup_populated_instr()
        self.assertFalse(has_component(instr, component_type="Arm"))

    def test_case_both(self):
        """
        check has_component when given both name and component type
        """
        instr = setup_populated_instr()
        self.assertTrue(has_component(instr, component_name="first_component",
                                      component_type="test_for_reading"))

    def test_case_both_not_found_name(self):
        """
        check has_component returns false when name not found but type is
        """
        instr = setup_populated_instr()
        self.assertFalse(has_component(instr, component_name="",
                                       component_type="test_for_reading"))

    def test_case_both_not_found_type(self):
        """
        check has_component returns false when type not found but name is
        """
        instr = setup_populated_instr()
        self.assertFalse(has_component(instr, component_name="first_component",
                                       component_type="Arm"))

    def test_parameter_found(self):
        """
        check has_parameter of given name
        """
        instr = setup_populated_instr()
        self.assertTrue(has_parameter(instr, parameter_name="theta"))

    def test_parameter_not_found(self):
        """
        check has_parameter returns false when of given name
        """
        instr = setup_populated_instr()
        self.assertFalse(has_parameter(instr, parameter_name="bogus"))

    def test_parameter_found_with_type(self):
        """
        check has_parameter of given name and type
        """
        instr = setup_populated_instr()
        self.assertTrue(has_parameter(instr, parameter_name="n_pulses", parameter_type="int"))

    def test_parameter_found_with_type_default(self):
        """
        check has_parameter of given name and type (using default double)
        """
        instr = setup_populated_instr()
        self.assertTrue(has_parameter(instr, parameter_name="theta", parameter_type="double"))

    def test_parameter_found_with_wrong_type(self):
        """
        check has_parameter of given name
        """
        instr = setup_populated_instr()
        self.assertFalse(has_parameter(instr, parameter_name="theta", parameter_type="int"))

    def test_all_parameters_set_not_set(self):
        """
        check all_parameters_set returns false when not all parameters are set of given name
        """
        instr = setup_populated_instr()
        self.assertFalse(all_parameters_set(instr))

    def test_all_parameters_set_actually_set(self):
        """
        check all_parameters_set returns false when all parameters are set of given name
        """
        instr = setup_populated_instr()

        instr.set_parameters(theta=37, n_pulses=3)
        self.assertTrue(all_parameters_set(instr))


import sys
import os
import threading

from collections import OrderedDict

import ipywidgets as widgets
from ipywidgets import GridBox
from IPython.display import display, clear_output

import matplotlib.pyplot as plt

from mcstasscript.interface.functions import name_search
from mcstasscript.helper.plot_helper import _plot_fig_ax

from mcstasscript.jb_interface.widget_helpers import HiddenPrints


class PlotInterface:
    """
    Class for providing plotting interface given McStasScript data
    """
    def __init__(self, data=None):
        """
        Initialize interface for exploring a dataset with plotting options

        Parameters
        ----------

        data: List of McStasData objects
            Optional to set the data, otherwise use set_data method
        """
        self.data = data

        # Variables related to monitor choice
        self.monitor_dropdown = None
        self.current_monitor = None

        # default plotting
        self.log_mode = None
        self.orders_of_mag = 300 # default value in McStasScript
        self.colormap = "jet"

        # Matplotlib objects
        self.fig = None
        self.ax = None
        self.colorbar_ax = None

    def set_data(self, data):
        """
        Set a new dataset for the interface, and updates the plot

        Parameters
        ----------

        data: List of McStasData objects
            New dataset that will be plotted
        """
        self.data = data
        self.monitor_dropdown.set_data(data)
        self.update_plot()

    def set_current_monitor(self, monitor):
        """
        Selects a new monitor to be plotted
        """
        self.current_monitor = monitor
        self.update_plot()

    def set_log_mode(self, log_mode):
        """
        Sets log mode for plotting, True or False
        """
        self.log_mode = bool(log_mode)
        self.update_plot()

    def set_orders_of_mag(self, orders_of_mag):
        """
        Sets orders_of_mag value for logarithmic plots
        """
        self.orders_of_mag = orders_of_mag
        self.update_plot()

    def set_colormap(self, colormap):
        """
        Choose colormap, has to be available in matplotlib
        """
        self.colormap = colormap
        self.update_plot()

    def new_plot(self):
        """
        Sets up original plot with fig, ax and ax for colorbar
        """
        # fig, ax = plt.subplots(constrained_layout=True, figsize=(6, 4))

        self.fig, (self.ax, self.colorbar_ax) = plt.subplots(ncols=2,
                                                             gridspec_kw={'width_ratios': [6, 1]},
                                                             tight_layout=True)

        self.fig.canvas.toolbar_position = 'bottom'

        plt.show()

        self.update_plot()

    def update_plot(self):
        """
        Updates the plot with current data, monitor and plot options

        Threading lock is used as this method is used in a threading context
        and can easily fail if new data is written while plotting. The lock
        prevents this from happening.
        """
        lock = threading.Lock()

        with lock:
            # Clear plot first
            self.ax.cla()
            #self.ax.xaxis.set_ticks([])
            #self.ax.yaxis.set_ticks([])
            self.colorbar_ax.cla()
            #self.colorbar_ax.xaxis.set_ticks([])
            #self.colorbar_ax.yaxis.set_ticks([])

            # Display message if not data can be plotted
            if self.data is None:
                self.ax.text(0.3, 0.5, "No data available yet")
                self.colorbar_ax.set_axis_off()
                self.ax.xaxis.set_ticks([])
                self.ax.yaxis.set_ticks([])
                return

            if len(self.data) == 0:
                self.ax.text(0.25, 0.5, "Simulation returned no data")
                self.colorbar_ax.set_axis_off()
                self.ax.xaxis.set_ticks([])
                self.ax.yaxis.set_ticks([])
                return

            if self.current_monitor is None:
                self.ax.text(0.3, 0.5, "Select a monitor to plot")
                self.colorbar_ax.set_axis_off()
                self.ax.xaxis.set_ticks([])
                self.ax.yaxis.set_ticks([])
                return

            # Get monitor and establish plot options
            monitor = name_search(self.current_monitor, self.data)
            plot_options = {"show_colorbar": True, "log": self.log_mode, "colormap": self.colormap}
            if self.orders_of_mag != "disabled":
                plot_options["orders_of_mag"] = self.orders_of_mag
            else:
                plot_options["orders_of_mag"] = 300 # Default value in McStasPlotOptions

            #print("Plotting with: ", plot_options)
            monitor.set_plot_options(**plot_options)
            with HiddenPrints():
                _plot_fig_ax(monitor, self.fig, self.ax, colorbar_axes=self.colorbar_ax)

            self.colorbar_ax.set_aspect(20)

            # Show colorbar if something is present, otherwise hide it
            if self.colorbar_ax.lines or self.colorbar_ax.collections:
                self.colorbar_ax.set_axis_on()
            else:
                self.colorbar_ax.set_axis_off()

            self.fig.canvas.draw()

    def show_interface(self):
        """
        Show the plot interface
        """
        # turn off automatic plotting
        plt.ioff()

        # Set up plot area
        output = widgets.Output()

        with output:
            self.new_plot()

        output.layout = widgets.Layout(width="75%")

        # could retrieve default plot options from data if given

        plot_control_list = []  # Keep all control widgets in this list
        plot_control_list.append(widgets.Label(value="Choose monitor"))

        # Set up dropdown list for monitor choice
        self.monitor_dropdown = MonitorDropdown(self.set_current_monitor)
        plot_control_list.append(self.monitor_dropdown.make_widget())

        plot_control_list.append(widgets.Label(value="Plot options"))

        # Set up checkbox for log plotting
        log_checkbox = LogCheckbox(self.log_mode, self.set_log_mode)
        plot_control_list.append(log_checkbox.make_widget())

        # Set up text field for orders of mag input
        log_orders_of_mag = OrdersOfMagField(self.set_orders_of_mag)
        plot_control_list.append(log_orders_of_mag.make_widget())

        # Set up dropdown box for colormap
        colormap_control = ColormapDropdown(self.set_colormap)
        plot_control_list.append(colormap_control.make_widget())

        plot_controls = widgets.VBox(plot_control_list,
                                     layout=widgets.Layout(width="25%"))

        # In case data is already supplied, set it
        if self.data is not None:
            self.set_data(self.data)

        return widgets.HBox([output, plot_controls])


class ColormapDropdown:
    """
    Class for controlling dropdown menus for colormaps
    """
    def __init__(self, set_colormap):
        """
        Controls colormap dropdown menus with given set_colormap function

        Creates dropdown widget and calls the given set_colormap function
        when the user updates the colormap choice.
        The colormap choice is given as two dropdown menus, one for selecting
        the category of colormap, and the other to select the actual colormap.

        The available colormaps are those supported in matplotlib

        Parameters
        ----------

        set_colormap : function
            Function called with colormap name as argument when changed

        """
        self.set_colormap = set_colormap
        self.colormap_widget = None

        # Default colormaps in matplotlib
        self.cmaps = OrderedDict()
        self.cmaps['Perceptually Uniform Sequential'] = [
            'viridis', 'plasma', 'inferno', 'magma', 'cividis']
        self.cmaps['Sequential'] = [
            'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
            'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
            'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']
        self.cmaps['Sequential (2)'] = [
            'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
            'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
            'hot', 'afmhot', 'gist_heat', 'copper']
        self.cmaps['Diverging'] = [
            'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
            'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']
        self.cmaps['Cyclic'] = ['twilight', 'twilight_shifted', 'hsv']
        self.cmaps['Qualitative'] = ['Pastel1', 'Pastel2', 'Paired', 'Accent',
                                'Dark2', 'Set1', 'Set2', 'Set3',
                                'tab10', 'tab20', 'tab20b', 'tab20c']
        self.cmaps['Miscellaneous'] = [
            'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
            'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg',
            'gist_rainbow', 'rainbow', 'jet', #'turbo',  # turbo reports errors
            'nipy_spectral', 'gist_ncar']

        self.categories = self.cmaps.keys()

    def make_widget(self):
        """
        Creates the widget and sets appropriate update functions

        The category dropdown will change options for the actual colormap
        dropdown menu through the update_cmap_options method.
        The colormap choice uses the update_cmap method and calls the
        update_function held in attributes.
        """

        header = widgets.Label(value="Colormap category")
        category = widgets.Dropdown(value="Miscellaneous", options=self.categories,
                                    layout=widgets.Layout(width="98%"))
        category.observe(self.update_cmap_options, "value")

        description = widgets.Label(value="Colormap", layout=widgets.Layout(width="39%"))
        default_options = self.cmaps[category.value]
        self.colormap_widget = widgets.Dropdown(value="jet", options=default_options,
                                                layout=widgets.Layout(width="58%"))
        self.colormap_widget.observe(self.update_cmap, "value")

        colormap = widgets.HBox([description, self.colormap_widget])

        return widgets.VBox([header, category, colormap])

    def update_cmap_options(self, change):
        """
        Updates the colormap options in the colormap widget
        """
        self.colormap_widget.options = self.cmaps[change.new]

    def update_cmap(self, change):
        """
        Updates the colormap with the set_colormap function in attributes
        """
        self.set_colormap(change.new)


class MonitorDropdown:
    """
    Class for creating monitor dropdown menu
    """
    def __init__(self, set_current_monitor):
        """
        Sets up MonitorDropdown menu with given update function

        Parameters
        ----------

        update_plot: function
            function called to update plot
        """

        self.set_current_monitor = set_current_monitor
        self.last_monitor = None
        self.data = None
        self.widget = None

    def set_data(self, data):
        """
        Updates the menu options given new data

        Parameters
        ----------

        data: McStasData list
            Data returned by McStasScript simulation
        """

        lock = threading.Lock()
        with lock:

            self.data = data

            if data is None:
                self.widget.options = []
                return

            monitor_names = []
            for data in self.data:

                # Ensure data names are unique
                original_name = data.name
                index = 1
                while data.name in monitor_names:
                    data.name = original_name + "_" + str(index)
                    index += 1

                monitor_names.append(data.name)

            self.widget.options = monitor_names

            # Go to the last set monitor if possible
            if self.last_monitor is not None:
                if self.last_monitor in self.widget.options:
                    self.set_current_monitor(self.last_monitor)

    def make_widget(self):
        """
        Builds the widget for the dropdown menu and links to update method

        """
        self.widget = widgets.Dropdown(layout=widgets.Layout(width="98%"))

        self.widget.observe(self.update, "value")

        return self.widget

    def update(self, change):
        """
        Calls update_function when dropdown menu is used

        Parameters
        ----------

        change: widget change
            state change of widget
        """
        # can do input sanitation here
        lock = threading.Lock()
        with lock:
            self.set_current_monitor(change.new)
            if change.new is not None:
                self.last_monitor = change.new


class LogCheckbox:
    """
    Class for widget with log mode checkbox
    """
    def __init__(self, log_mode, set_log_mode):
        """
        Sets up checkbox with log mode, takes initial mode and update function

        Parameters
        ----------

        log_mode : bool
            Initial state of log checkbox

        set_log_mode : function
            Function which will be called with new log_mode
        """
        self.log_mode = log_mode
        self.set_log_mode = set_log_mode

    def make_widget(self):
        """
        Creates the actual checkbox widget along with descriptive text
        """
        description_layout = widgets.Layout(width='250px', height='32px',
                                            display="flex",
                                            justify_content="flex-start")
        description_log = widgets.Label(value="Log plot", layout=description_layout)

        log_check = widgets.Checkbox(
            value=self.log_mode,
            description='',
            disabled=False,
            indent=False,
            layout=widgets.Layout(width='70px', height='32px')
        )
        log_check.observe(self.update, "value")

        return widgets.HBox([description_log, log_check])

    def update(self, change):
        """
        Calls given update function with new log_mode

        Parameters
        ----------

        change : change object
            Change object from widget interaction
        """
        self.set_log_mode(change.new)


class OrdersOfMagField:
    """
    Class for handling orders_of_mag widget
    """
    def __init__(self, set_orders_of_mag):
        """
        Widget for entering orders_of_mag used in log plotting

        Orders_of_mag parameter controls how many orders of magnitude are
        plotted when showing log scale, counting from the highest value and
        down. The object needs an update function orders_of_mag.
        The widget text field starts with the text "disabled" which
        corresponds to the default value of 300 orders of magnitude.

        Parameters
        ----------

        set_orders_of_mag : function
            Function for updating orders_of_mag value
        """
        self.set_orders_of_mag = set_orders_of_mag

    def make_widget(self):
        """
        Creates actual widget with descriptive text
        """
        description_layout = widgets.Layout(width='250px', height='32px',
                                            display="flex",
                                            justify_content="flex-start")
        description_orders = widgets.Label(value="Orders of magnitude", layout=description_layout)
        textbox = widgets.Text(value=str("disabled"),
                               layout=widgets.Layout(width='70px', height='32px'))
        textbox.observe(self.update, "value")

        return widgets.HBox([description_orders, textbox])

    def update(self, change):
        """
        Update orders_of_mag and replot

        Parameters
        ----------

        change: widget change
            state change of widget
        """
        if change.new == "disabled":
            self.set_orders_of_mag(change.new)
            return

        try:
            float(change.new)
        except ValueError:
            return

        self.set_orders_of_mag(float(change.new))



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import BoundaryNorm

from mcstasscript.instrument_diagnostics.diagnostics_instrument import DiagnosticsInstrument
from mcstasscript.interface.functions import name_search
from mcstasscript.data.data import McStasDataBinned
from mcstasscript.helper.plot_helper import _plot_fig_ax

class IntensityDiagnostics(DiagnosticsInstrument):
    def __init__(self, instr):
        super().__init__(instr)

        self.data = None
        self.data_dim = None
        self.monitors = None

    def add_monitor(self, before, options):
        name = "I_before_" + before.name
        mon = self.instr.add_component(name, "Monitor_nD", before=before)

        mon.set_parameters(restore_neutron=1,
                           xwidth=100, yheight=100,
                           options=options,
                           filename='"' + name + ".diag" + '"')

        mon.set_AT(before.AT_data, RELATIVE=before.AT_reference)
        if before.ROTATED_specified:
            mon.set_ROTATED(before.ROTATED_data, RELATIVE=before.ROTATED_reference)

        return name

    def run_general(self, variable=None, limits=None):#, start=None, end=None):
        self.reset_instr()
        self.remove_previous_use()

        if limits is None:
            limit_string = "auto"
        else:
            if not isinstance(limits, list) or len(limits) != 2:
                raise TypeError("limits has to be a list of length 2.")
            limit_string = f"limits=[{limits[0]},{limits[1]}]"

        if variable is None:
            options = f'"square boarders intensity"'
            self.data_dim = 0
        else:
            options = f'"square {variable} {limit_string} bins=300"'
            self.data_dim = 1

        self.monitors = []
        for comp in self.component_list[1:]:
            mon_name = self.add_monitor(before=comp, options=options)
            self.monitors.append((mon_name, comp.name))

        self.correct_target_index()

        self.data = self.instr.backengine()

    def run(self):#, start=None, end=None):
        self.reset_instr()
        self.remove_previous_use()

        options = f'"square boarders intensity"'
        self.monitors = []
        for comp in self.component_list[1:]:
            mon_name = self.add_monitor(before=comp, options=options)
            self.monitors.append((mon_name, comp.name))

        self.correct_target_index()

        self.data = self.instr.backengine()

    def plot(self, figsize=None, ax=None, fig=None, show_comp_names=True,
             y_tick_positions=None, ylimits=None):

        if self.data_dim == 0:
            self.plot_0D(figsize=figsize, ax=ax, fig=fig,
                         show_comp_names=show_comp_names,
                         y_tick_positions=y_tick_positions,
                         ylimits=ylimits)
        elif self.data_dim == 1:
            self.plot_1D(figsize=figsize, ax=ax, fig=fig,
                         show_comp_names=show_comp_names,
                         y_tick_positions=y_tick_positions,
                         ylimits=ylimits)

    def plot_1D(self, figsize=None, ax=None, fig=None, show_comp_names=True,
                y_tick_positions=None, ylimits=None):

        if figsize is None:
            figsize = (8, len(self.component_list) / 5 + 1)

        data_sets = []
        for I_monitor_name, component_name in self.monitors:
            mon_data = name_search(I_monitor_name, self.data)
            intensity = mon_data.Intensity
            axis = mon_data.xaxis
            data_sets.append({"name": component_name, "I": intensity, "axis": axis})

        if y_tick_positions is None:
            y_positions = np.linspace(1, len(data_sets), len(data_sets))
            y_sep = y_positions[1] - y_positions[0]
            y_positions = np.append(y_positions, y_positions[-1] + y_sep)
        else:
            #y_tick_positions.reverse()
            y_positions = y_tick_positions

        # Find min and max for axis
        for index, data_set in enumerate(data_sets):
            if index == 0:
                max_axis = max(data_set["axis"])
                min_axis = min(data_set["axis"])
            else:
                max_axis = max(max(data_set["axis"]), max_axis)
                min_axis = min(min(data_set["axis"]), min_axis)

        # New axis from min to max
        n_bins = 300
        axis = np.linspace(min_axis, max_axis, n_bins)
        sep = axis[1] - axis[0]
        bin_axis = np.append(axis - 0.5 * sep, axis[-1] + sep)
        intensities = np.zeros((len(data_sets), len(axis)))

        for index, data_set in enumerate(data_sets):
            new_bins = np.digitize(data_set["axis"], bin_axis)
            boarder_bins = np.where(new_bins == n_bins)
            new_bins[boarder_bins] -= 1
            intensities[index, new_bins] += data_set["I"]

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)

        component_names = [x.name for x in self.original_instr.make_component_subset()]
        component_names.reverse()
        if not show_comp_names:
            component_names = [""] * len(component_names)

        metadata = mon_data.metadata
        metadata.dimension = [len(axis), len(y_positions) - 1]
        metadata.limits = [0, 0, 0, 0]
        metadata.limits[0] = min(axis)
        metadata.limits[1] = max(axis)
        metadata.limits[2] = min(y_positions)
        metadata.limits[3] = max(y_positions)

        display_data = np.flip(intensities, 0)
        data = McStasDataBinned(metadata, display_data, np.zeros((1, 1)), np.zeros((1, 1)))

        data.set_plot_options(show_colorbar=False, log=True, orders_of_mag=5)
        data.set_ylabel("")

        _plot_fig_ax(data, fig, ax)

        ax.set_yticks(y_positions)
        ax.set_yticklabels(component_names, fontsize=18)
        ax.set_xlabel(mon_data.metadata.xlabel, fontsize=18)

    def plot_0D(self, figsize=None, ax=None, fig=None, show_comp_names=True,
                y_tick_positions=None, ylimits=None):

        if figsize is None:
            figsize = (8, len(self.component_list)/5 + 1)

        intensities = []
        ray_counts = []
        component_names = []
        indicies = []

        index = 0
        for I_monitor_name, component_name in self.monitors:
            mon_data = name_search(I_monitor_name, self.data)
            values = mon_data.metadata.info["values"].split()
            # values contain strings of: Intensity, Error, Ncount
            intensities.append(float(values[0]))
            ray_counts.append(float(values[2]))
            component_names.append(component_name)
            indicies.append(index)
            index += 1

        # Extend with the last one
        intensities.append(intensities[-1])
        ray_counts.append(ray_counts[-1])
        indicies.append(index)
        component_names = [self.component_list[0].name] + component_names

        if not show_comp_names:
            component_names = [""] * len(component_names)

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)

        intensities.reverse()
        ray_counts.reverse()
        component_names.reverse()

        if y_tick_positions is None:
            y_positions = indicies
        else:
            y_tick_positions.reverse()
            y_positions = y_tick_positions

        # Ensure x scale for intensity and n count share same tick marks
        I_limits, N_limits = common_range_limits(intensities, ray_counts)

        ax.step(intensities, y_positions, where="post", color="k", zorder=3.5)
        ax.set_yticks(y_positions)
        ax.set_yticklabels(component_names, fontsize=18)
        ax.set_xlabel("Intensity [n/s]", fontsize=18, color="k")
        ax.set_xscale("log", nonpositive='clip')
        ax.xaxis.set_tick_params(labelsize=16)
        ax.set_xlim(I_limits)

        if ylimits is None:
            ax.set_ylim([-0.5, index + 0.5])
        else:
            ax.set_ylim(ylimits)

        ax.grid(True)

        ax2 = ax.twiny()
        ax2.step(ray_counts, y_positions, where="post", color="g", linestyle="--", zorder=3.6)
        ax2.set_xlabel("Ray count", fontsize=18, color="g")
        ax2.set_xscale("log", nonpositive='clip')
        ax2.xaxis.set_tick_params(labelsize=16, colors="g")
        ax2.set_xlim(N_limits)

def common_range_limits(data_I, data_N):
    # Convert to numpy
    data_I = np.array(data_I)
    data_I_nonzero = data_I[np.nonzero(data_I)]
    data_N = np.array(data_N)
    data_N_nonzero = data_N[np.nonzero(data_N)]

    max_I = max(data_I_nonzero)
    min_I = min(data_I_nonzero)
    I_orders_of_mag = np.log10(max_I) - np.log10(min_I)

    max_N = max(data_N_nonzero)
    min_N = min(data_N_nonzero)
    N_orders_of_mag = np.log10(max_N) - np.log10(min_N)

    I_is_largest = I_orders_of_mag > N_orders_of_mag

    if I_is_largest:
        # Use intensity scale
        max_large = max_I
        min_large = min_I
        max_small = max_N
    else:
        max_large = max_N
        min_large = min_N
        max_small = max_I

    log_extra = 0.1  # Extra scale to avoid having data just at the edge

    # Round I scale up
    log_large_scale_max = np.ceil(np.log10(max_large)) + log_extra
    log_large_scale_min = np.floor(np.log10(min_large))
    large_limits = [10 ** log_large_scale_min, 10 ** log_large_scale_max]

    # Find how many orders of mag this cover
    large_orders_of_mag = log_large_scale_max - log_large_scale_min

    # Apply same to ray count
    log_small_scale_max = np.ceil(np.log10(max_small)) + log_extra
    log_small_scale_min = log_small_scale_max - large_orders_of_mag
    small_limits = [10 ** log_small_scale_min, 10 ** log_small_scale_max]

    if I_is_largest:
        I_limits = large_limits
        N_limits = small_limits
    else:
        I_limits = small_limits
        N_limits = large_limits

    return I_limits, N_limits


class Arrow:
    """
    Helper class for creating arrows with connections
    """
    def __init__(self, origin, target, lane=None, kind=None, description=None):
        """
        Arrow object with origin Component_box and target component_box
        """
        self.origin = origin
        self.target = target
        self.kind = kind
        self.description=description

        self.lane = lane
        self.sub_lane = 0
        self.lane_width = None
        self.lane_offset = 0

        self.box_offset_origin = 0
        self.box_offset_target = 0

        self.origin_linestyle = "-"
        self.color = "blue"
        self.alpha = 1.0
        self.arrow_width = 0.003
        self.arrow_length = 0.01
        self.connection = False

        self.origin_congested = False
        self.target_congested = False

    def set_lane(self, lane):
        self.lane = lane

    def set_sub_lane(self, lane):
        self.sub_lane = lane

    def set_lane_width(self, lane_width):
        self.lane_width = lane_width

    def set_lane_offset(self, offset):
        self.lane_offset = offset

    def get_lane_value(self):
        return self.lane_width * (self.lane + self.sub_lane * self.lane_offset)

    def set_connection(self, value):
        self.connection = value

    def set_origin_congested(self, value):
        self.origin_congested = value

    def set_target_congested(self, value):
        self.target_congested = value

    def set_box_offset_origin(self, value):
        """
        In units of box height
        """
        self.box_offset_origin = value

    def get_box_offset_origin(self):
        return self.origin.box_height * self.box_offset_origin

    def set_box_offset_target(self, value):
        """
        In units of box height
        """
        self.box_offset_target = value

    def get_box_offset_target(self):
        return self.target.box_height * self.box_offset_target

    def set_arrow_width(self, value):
        self.arrow_width = value

    def set_arrow_length(self, value):
        self.arrow_length = value

    def set_linestyle(self, value):
        self.origin_linestyle = value

    def set_alpha(self, value):
        self.alpha = value

    def plot_left_side(self, ax):
        # stop a bit before the lane and make a little triangle
        triangle_leg_x = 0.25 * self.arrow_length
        triangle_leg_y = 2.0 * self.arrow_width

        # box to lane
        origin_x = self.origin.get_text_start()
        origin_y = self.origin.position_y + self.get_box_offset_origin()

        origin_lane_x = origin_x - self.get_lane_value()
        origin_lane_y = origin_y

        entry_point_x = origin_lane_x + triangle_leg_x

        x_points = [origin_x, entry_point_x]
        y_points = [origin_y, origin_lane_y]

        # lane
        target_x = self.target.get_text_start()
        target_y = self.target.position_y + self.get_box_offset_target()

        target_lane_x = target_x - self.get_lane_value()
        target_lane_y = target_y

        exit_point_x = target_lane_x + triangle_leg_x

        # Plot lane but ease in and out of the lane with a small triangle
        x_points += [entry_point_x, origin_lane_x, origin_lane_x, target_lane_x, origin_lane_x, exit_point_x]
        y_points += [origin_lane_y, origin_lane_y + triangle_leg_y, origin_lane_y + triangle_leg_y,
                     target_lane_y - triangle_leg_y, target_lane_y - triangle_leg_y, target_lane_y]
        ax.plot(x_points, y_points, color=self.color, linestyle=self.origin_linestyle)

        # lane to box
        arrow_length = self.target.get_text_start() - target_lane_x - triangle_leg_x
        ax.arrow(x=target_lane_x + triangle_leg_x, y=target_lane_y,
                 dx=arrow_length, dy=0,
                 color=self.color, length_includes_head=True,
                 width=self.arrow_width, head_width=5.0*self.arrow_width,
                 head_length=self.arrow_length)

    def plot_right_side(self, ax, text_end):
        # stop a bit before the lane and make a little triangle
        triangle_leg_x = 0.25 * self.arrow_length
        triangle_leg_y = 2.0 * self.arrow_width

        # origin to lane
        origin_x = self.origin.get_text_end()
        origin_y = self.origin.position_y + self.get_box_offset_origin()

        origin_lane_x = text_end + self.get_lane_value()
        origin_lane_y = origin_y

        entry_point_x = origin_lane_x - triangle_leg_x

        x_points = [origin_x, entry_point_x]
        y_points = [origin_y, origin_lane_y]

        """
        ax.plot([origin_x, entry_point_x], [origin_y, origin_lane_y],
                color=self.color, linestyle=self.origin_linestyle, alpha=self.alpha)
        """

        if self.description is not None and not self.origin_congested:
            bbox = dict(boxstyle="round", facecolor="white", edgecolor="white", alpha=0.85)
            ax.text(0.5*(origin_x + origin_lane_x), origin_lane_y,
                    self.description, ha="center", va="center", bbox=bbox)

        # lane
        target_y = self.target.position_y + self.get_box_offset_target()

        target_lane_x = text_end + self.get_lane_value()
        target_lane_y = target_y

        exit_point_x = target_lane_x - triangle_leg_x

        if target_lane_y < origin_lane_y:
            leg_dir = 1
        else:
            leg_dir = -1

        triangle_leg_y_dir = leg_dir * triangle_leg_y

        # Plot lane but ease in and out of it with a small triangle
        x_points += [entry_point_x, target_lane_x, origin_lane_x, target_lane_x, origin_lane_x, exit_point_x]
        y_points += [origin_lane_y, origin_lane_y - triangle_leg_y_dir, origin_lane_y - triangle_leg_y_dir,
                     target_lane_y + triangle_leg_y_dir, target_lane_y + triangle_leg_y_dir, target_lane_y]
        ax.plot(x_points, y_points, color=self.color, alpha=self.alpha, linestyle=self.origin_linestyle)

        # target
        if self.connection:
            mid_point = 0.5*(target_lane_x + self.target.get_text_end())
            ax.plot([target_lane_x - triangle_leg_x, self.target.get_text_end()],
                    [target_lane_y, target_lane_y], color=self.color, alpha=self.alpha)
        else:
            arrow_length = target_lane_x - self.target.get_text_end() - triangle_leg_x
            mid_point = target_lane_x - 0.5*arrow_length
            ax.arrow(x=target_lane_x - triangle_leg_x, y=target_lane_y,
                     dx=-arrow_length, dy=0,
                     color=self.color, length_includes_head=True,
                     width=self.arrow_width, head_width=5.0*self.arrow_width,
                     head_length=self.arrow_length, alpha=self.alpha)

        if self.description is not None and not self.target_congested:
            bbox = dict(boxstyle="round", facecolor="white", edgecolor="white", alpha=0.85)
            ax.text(mid_point, target_lane_y,
                    self.description, ha="center", va="center", bbox=bbox)


# Version number
from ._version import __version__

from .interface.instr import McStas_instr
from .interface.instr import McXtrace_instr

from .interface.functions import load_data
from .interface.functions import load_metadata
from .interface.functions import load_monitor
from .interface.functions import name_plot_options
from .interface.functions import name_search
from .interface.functions import Configurator

from .interface.plotter import make_animation
from .interface.plotter import make_plot
from .interface.plotter import make_sub_plot

from .interface.reader import McStas_file

from .tools.cryostat_builder import Cryostat
from .tools.instrument_checker import has_component, has_parameter, all_parameters_set

from .instrument_diagnostics.beam_diagnostics import BeamDiagnostics as Diagnostics



from mcstasscript.instrument_diagram.connections import ConnectionList
from mcstasscript.instrument_diagram.arrow import Arrow


def generate_JUMP_arrows(components, component_box_dict, box_names, color=None):
    """
    Generate Arrow objects related to the JUMP keyword
    """
    connections = ConnectionList()

    for component in components:
        if component.JUMP != "":
            jump_reference = component.JUMP.split(" ")[0]
            if jump_reference not in component_box_dict:
                raise ValueError("JUMP reference: " + str(jump_reference)
                                 + " not found.")

            origin = component_box_dict[component.name]
            connections.add(origin, component_box_dict[jump_reference])

    connections.distribute_lane_numbers(box_names=box_names)

    arrows = []
    for connection in connections.get_connections():
        origin = connection.origin
        target = connection.target
        lane = connection.lane_number + 2  # Make room for group name

        arrow = Arrow(origin, target, lane=lane, kind="JUMP", description="JUMP")
        arrow.set_sub_lane(1)
        if color is None:
            arrow.color = "black"
        else:
            arrow.color = color

        arrows.append(arrow)

    return arrows



import os
import io
import unittest
import unittest.mock
import ipywidgets as widgets

import matplotlib
matplotlib.use('Agg')  # Set the backend to Agg

from mcstasscript.jb_interface.simulation_interface import SimInterface
from mcstasscript.jb_interface.simulation_interface import ParameterWidget
from mcstasscript.jb_interface.widget_helpers import parameter_has_default
from mcstasscript.jb_interface.widget_helpers import get_parameter_default
from mcstasscript.interface.instr import McStas_instr
from mcstasscript.interface.instr import McXtrace_instr


def setup_instr_root_path_McStas():
    """
    Sets up a neutron instrument with root package_path
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McStas_instr("test_instrument", package_path="/")

    os.chdir(current_work_dir)

    return instrument

def setup_populated_instr_McStas():
    """
    Sets up a neutron instrument with some features used and three components
    """
    instr = setup_instr_root_path_McStas()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    instr.add_component("first_component", "test_for_reading")
    instr.add_component("second_component", "test_for_reading")
    instr.add_component("third_component", "test_for_reading")

    return instr

def setup_instr_root_path_McXtrace():
    """
    Sets up a neutron instrument with root package_path
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McStas_instr("test_instrument", package_path="/")

    os.chdir(current_work_dir)

    return instrument

def setup_populated_instr_McXtrace():
    """
    Sets up a neutron instrument with some features used and three components
    """
    instr = setup_instr_root_path_McXtrace()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    instr.add_component("first_component", "test_for_reading")
    instr.add_component("second_component", "test_for_reading")
    instr.add_component("third_component", "test_for_reading")

    return instr


class FakeChange:
    def __init__(self, new=None, old=None, name=None):
        self.new = new
        self.old = old
        self.name = name


class TestSimulationInterface(unittest.TestCase):
    """
    Tests of simulation interface
    """
    def test_initialization_McStas(self):
        """
        Checking that interface can initialize from instrument and retrieve the
        parameters that has a default value.
        """

        sim_interface = SimInterface(setup_populated_instr_McStas())
        self.assertEqual(sim_interface.parameters["has_default"], 37)

    def test_show_interface_McStas(self):
        """
        Ensure that show_interface runs without errors and returns widget
        """

        sim_interface = SimInterface(setup_populated_instr_McStas())
        widget = sim_interface.show_interface()

        self.assertIsInstance(widget, widgets.widgets.widget_box.VBox)

    def test_initialization_McXtrace(self):
        """
        Checking that interface can initialize from instrument and retrieve the
        parameters that has a default value.
        """

        sim_interface = SimInterface(setup_populated_instr_McXtrace())
        self.assertEqual(sim_interface.parameters["has_default"], 37)

    def test_show_interface_McXtrace(self):
        """
        Ensure that show_interface runs without errors and returns widget
        """

        sim_interface = SimInterface(setup_populated_instr_McXtrace())
        widget = sim_interface.show_interface()

        self.assertIsInstance(widget, widgets.widgets.widget_box.VBox)

    def test_update_ncount(self):
        """
        Check ncount can be set correctly
        """
        sim_interface = SimInterface(setup_populated_instr_McStas())
        sim_interface.show_interface()

        fake_change = FakeChange(new=100)
        sim_interface.update_ncount(fake_change)
        self.assertEqual(sim_interface.ncount, 100)

    def test_update_mpi(self):
        """
        Check mpi can be set correctly
        """
        sim_interface = SimInterface(setup_populated_instr_McStas())
        sim_interface.show_interface()

        fake_change = FakeChange(new=3)
        sim_interface.update_mpi(fake_change)
        self.assertEqual(sim_interface.mpi, 3)

        # Check input that wouldn't work is ignored
        fake_change = FakeChange(new="wrong input")
        sim_interface.update_mpi(fake_change)
        self.assertEqual(sim_interface.mpi, 3)

    def test_ParameterWidget(self):
        """
        Test that ParameterWidgets are initialized correctly

        This code is part of the initialization for the simulation interface,
        yet it is useful to have as its own test as if it fails its clear where
        the overall problem is.
        """

        instrument = setup_populated_instr_McStas()

        instrument.add_parameter("string", "choice", options=["A", "B", "Long"])

        parameters = {}
        # get default parameters from instrument
        for parameter in instrument.parameters:
            if parameter_has_default(parameter):
                parameters[parameter.name] = get_parameter_default(parameter)

        parameter_widgets = []
        parameterwidget_objects = []
        for parameter in instrument.parameters:
            par_widget = ParameterWidget(parameter, parameters)
            parameterwidget_objects.append(par_widget)
            parameter_widgets.append(par_widget.make_widget())

        self.assertEqual(parameterwidget_objects[0].name, "theta")
        self.assertEqual(parameterwidget_objects[0].default_value, None)
        # Parameter does not exist in parameter dict yet
        with self.assertRaises(KeyError):
            parameters[parameterwidget_objects[0].name]

        change = FakeChange(new=222)
        parameterwidget_objects[0].update(change)
        self.assertEqual(parameters[parameterwidget_objects[0].name], 222)

        self.assertEqual(parameterwidget_objects[1].name, "has_default")
        self.assertEqual(parameterwidget_objects[1].default_value, 37)
        self.assertEqual(parameters[parameterwidget_objects[1].name], 37)

        change = FakeChange(new=227)
        parameterwidget_objects[1].update(change)
        self.assertEqual(parameters[parameterwidget_objects[1].name], 227)

        self.assertEqual(parameterwidget_objects[2].name, "choice")
        self.assertEqual(parameterwidget_objects[2].default_value, None)
        with self.assertRaises(KeyError):
            parameters[parameterwidget_objects[2].name]

        change = FakeChange(new="test")
        parameterwidget_objects[2].update(change) # Should add necessary extra quotation marks
        self.assertEqual(parameters[parameterwidget_objects[2].name], "\"test\"")




import sys
import os
import numpy as np
import threading
import copy

import ipywidgets as widgets
from IPython.display import display

import matplotlib.pyplot as plt

from mcstasscript.interface.functions import name_search
from mcstasscript.interface import plotter
from mcstasscript.jb_interface import plot_interface
from mcstasscript.jb_interface.widget_helpers import HiddenPrints
from mcstasscript.jb_interface.widget_helpers import parameter_has_default
from mcstasscript.jb_interface.widget_helpers import get_parameter_default


class SimInterface:
    """
    Class for setting up widget that controls McStasScript instrument and plot
    """
    def __init__(self, instrument):
        """
        Sets up widget where the user can input instrument parameters, run the
        simulation, see plotted results and adjust the plots.

        The parameters of the instrument model are displayed with name, default
        value and comment. Can be adjusted with free text.

        A run button starts a simulation, and basic settings can be adjusted.

        A dropdown menu is available for selecting what monitor to view results
        from, and basic settings related to the plot can be adjusted.

        Show the interface with the show_interface method.

        Parameters
        ----------

        instrument: McStas_instr or McXtrace_instr
            instrument for which a widget should be created
        """

        self.instrument = instrument

        self.plot_interface = None

        self.run_button = None
        self.live_widget = None
        self.progress_bar = None
        self.sim_steps = 5

        self.ncount = "1E6"
        self.mpi = "disabled"
        self.last_mpi_on = None

        self.thread_data = None
        self.thread = None
        self.run_arguments = None

        self.parameters = {}
        # get default parameters from instrument
        for parameter in self.instrument.parameters:
            if parameter_has_default(parameter):
                self.parameters[parameter.name] = get_parameter_default(parameter)

    def make_parameter_widgets(self):
        """
        Creates widgets for parameters using dedicated class ParameterWidget

        returns widget including all parameters
        """
        parameter_widgets = []
        for parameter in self.instrument.parameters:
            par_widget = ParameterWidget(parameter, self.parameters)
            parameter_widgets.append(par_widget.make_widget())

        return widgets.VBox(parameter_widgets)

    def run_simulation_thread(self, change):
        """
        Runs simulation as thread, allowing user to update plots simultaneously

        The use of this method has caused crashes, temporarily circumvented by
        calling run_simulation_live on button instead. Now plots can now be
        updated while a simulation is running.

        Parameters
        ----------

        change: widget change
            Not used
        """

        thread = threading.Thread(target=self.run_simulation_live, args=[1])
        thread.start()

    def run_simulation_live(self, change):
        """
        Performs the simulation with current parameters and settings.

        When live mode is used, updates plot as more data is added.

        Changes icon on button to hourglass while simulation is running, then
        returns to calculator icon.
        """

        lock = threading.Lock()

        if self.live_widget.value:
            sim_parts = self.sim_steps
            #self.plot_interface.set_data(None)
        else:
            sim_parts = 1

        part_ncount = int(float(self.ncount)/sim_parts)

        run_arguments = {"output_path": "interface_" + self.instrument.name,
                         "increment_folder_name": True,
                         #"parameters": self.parameters,
                         "ncount": part_ncount}
        if self.mpi != "disabled":
            run_arguments["mpi"] = self.mpi
            mpi_on = True
        else:
            mpi_on = False

        # McStas does not recognize if the instrument was compiled with or without mpi
        # Ensure it is compiled when switching to and from mpi
        # This also ensures the instrument is compiled at the first run
        if mpi_on == self.last_mpi_on:
            run_arguments["force_compile"] = False
        else:
            run_arguments["force_compile"] = True

        self.last_mpi_on = mpi_on

        self.run_button.icon = "hourglass"
        #print("Running with:", run_arguments)

        if self.live_widget.value:
            self.progress_bar.layout.visibility = 'visible'
        else:
            self.progress_bar.layout.visibility = 'hidden'

        self.progress_bar.value = 0
        plot_data = None
        for index in range(sim_parts):
            self.instrument.settings(**run_arguments)
            self.instrument.set_parameters(self.parameters)
            try:
                with HiddenPrints():
                    self.instrument.backengine()
            except NameError:
                print("McStas run failed.")
                data = []

            with lock:
                self.progress_bar.value = index + 1
                data = self.instrument.output.get_data()["data"]

                if data is not None:
                    if plot_data is None:
                        plot_data = data
                    else:
                        add_data(plot_data, data)

                    sent_data = copy.deepcopy(plot_data)
                    # This happens in a thread, maybe it should be in Main?
                    self.plot_interface.set_data(sent_data)

        self.run_button.icon = "calculator"

    def make_run_button(self):
        """
        Creates a run button which perform the simulation
        """
        button = widgets.Button(
            description='Run',
            disabled=False,
            button_style='',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Runs the simulation with current parameters',
            icon='calculator'  # (FontAwesome names without the `fa-` prefix)
        )
        #button.on_click(self.run_simulation_thread)
        button.on_click(self.run_simulation_live)

        return button

    def make_ncount_field(self):
        """
        Creates field for ncount, links to update_ncount

        The field supports scientific notation
        """
        description_layout = widgets.Layout(width='70px', height='32px',
                                            display="flex",
                                            justify_content="flex-end")
        description = widgets.Label(value="ncount", layout=description_layout)
        textbox = widgets.Text(value=str(self.ncount), layout=widgets.Layout(width='100px', height='36px'))
        textbox.observe(self.update_ncount, "value")

        return widgets.HBox([description, textbox])

    def update_ncount(self, change):
        """
        Updates ncount variable from textbox input

        Only updates when usable input is entered. Supports scientific
        notation in input through conversion to float

        Parameters
        ----------

        change: widget change
            state change of widget
        """
        try:
            self.ncount = int(float(change.new))
        except ValueError:
            pass

    def make_mpi_field(self):
        """
        Creates field for mpi, links to update_mpi
        """
        description_layout = widgets.Layout(width='40px', height='32px',
                                            display="flex",
                                            justify_content="flex-end")
        description = widgets.Label(value="mpi", layout=description_layout)
        textbox = widgets.Text(value=str(self.mpi), layout=widgets.Layout(width='70px', height='36px'))
        textbox.observe(self.update_mpi, "value")

        return widgets.HBox([description, textbox])

    def update_mpi(self, change):
        """
        Updates mpi value when integer or the word 'disabled' is given

        Parameters
        ----------

        change: widget change
            state change of widget
        """
        if change.new == "disabled":
            self.mpi = change.new

        try:
            self.mpi = int(change.new)
        except ValueError:
            pass

    def make_live_checkmark(self):
        """
        Makes widget for choosing live simulations on / off
        """
        widget = widgets.Checkbox(value=False, description="Live results")
        widget.layout.visibility = "hidden"

        return widget

    def make_progress_bar(self):
        """
        Makes a progress bar for live simulations
        """
        widget = widgets.IntProgress(value=0, min=0, max=self.sim_steps,
                                     description="Sim progress",
                                     orientation="horizontal")

        return widget

    def get_data(self):
        """
        Returns last data set from this interface
        """
        if self.plot_interface is None:
            print("No widget interface initialized, use show_interface method.")
            return []

        if self.plot_interface.data is None:
            print("No run has been performed with the interface widget yet")
            return []

        return self.plot_interface.data

    def show_interface(self):
        """
        Builds and shows widget interface
        """

        # Make parameter controls
        parameter_widgets = self.make_parameter_widgets()

        # Make simulation controls
        self.live_widget = self.make_live_checkmark()
        self.progress_bar = self.make_progress_bar()
        self.progress_bar.layout.visibility = "hidden"
        self.run_button = self.make_run_button()
        ncount_field = self.make_ncount_field()
        mpi_field = self.make_mpi_field()

        simulation_widget = widgets.HBox([self.run_button, ncount_field, mpi_field,
                                          self.live_widget, self.progress_bar])
                                         #layout=widgets.Layout(border="solid"))

        self.plot_interface = plot_interface.PlotInterface()
        plot_widget = self.plot_interface.show_interface()

        return widgets.VBox([parameter_widgets, simulation_widget, plot_widget])


class ParameterWidget:
    """
    Widget for parameter object from McStasScript instrument
    """
    def __init__(self, parameter, parameters):
        """
        Describes a widget for a parameter object given all parameters

        When no options are given in ParameterVariable object, the widget will
        be a textfield where the user can input the value. If the options
        attribute is used, the widget will be a dropdown menu with available
        options. The make_widget method returns the widget, and the update
        function is called whenever the user interacts with the widget.

        The widget shows parameter name, the interactive widget and a comment

        Parameters
        ----------

        parameter: McStasScript ParameterVariable object
            The parameter this widget should represent

        parameters: dict of McStasScript ParameterVariable objects
            Dict with all parameter objects of the instrument
        """

        self.parameter = parameter
        self.parameters = parameters

        if parameter_has_default(parameter):
            self.default_value = get_parameter_default(parameter)
        else:
            self.default_value = None

        self.name = parameter.name
        self.comment = parameter.comment

    def make_widget(self):
        """
        Returns widget with parameter name, interactive widget and comment
        """
        label = widgets.Label(value=self.name,
                              layout=widgets.Layout(width='15%', height='32px'))
        if len(self.parameter.get_options()) > 0:
            par_widget = widgets.Dropdown(options=self.parameter.get_options(),
                                          layout=widgets.Layout(width='10%', height='32px'))
            if self.default_value is not None:
                if self.default_value in self.parameter.get_options():
                    par_widget.value = self.default_value

                if isinstance(self.default_value, str):

                    if self.default_value.strip("'") in self.parameter.get_options():
                        par_widget.value = self.default_value.strip("'")
                    elif self.default_value.strip('"') in self.parameter.get_options():
                        par_widget.value = self.default_value.strip('"')

                if par_widget.value is None:
                    print(self.parameter.get_options())
                    raise KeyError("default value not found in options for parameter: "
                                   + str(self.parameter.name))

        else:
            par_widget = widgets.Text(value=str(self.default_value),
                                      layout=widgets.Layout(width='10%', height='32px'))
        comment = widgets.Label(value=self.comment,
                                layout=widgets.Layout(width='75%', height='32px'))

        par_widget.observe(self.update, "value")

        return widgets.HBox([label, par_widget, comment])

    def update(self, change):
        """
        Update function called whenever the user updates the widget

        When strings parameters are used, this function adds the necessary
        quotation marks if none are provided.
        """
        new_value = change.new
        if self.parameter.type == "string":
            if type(new_value) is str:
                if not (new_value[0] == '"' or new_value[0] == "'"):
                    new_value = '"' + new_value + '"'
        else:

            try:
                new_value = float(new_value)
            except:
                return

        self.parameters[self.name] = new_value


def add_data(initial, new_data):
    """
    Method for adding new data to a data set

    Updates Intensity, Error and Ncount

    Updates all data except metadata info
    """

    for monitor in initial:
        ref_ncount = float(monitor.metadata.info["Ncount"])

        new_monitor = name_search(monitor.name, new_data)
        new_ncount = float(new_monitor.metadata.info["Ncount"])

        total_ncount = ref_ncount + new_ncount

        scale_old = ref_ncount / total_ncount
        scale_new = new_ncount / total_ncount

        monitor.Intensity = scale_old*monitor.Intensity + scale_new*new_monitor.Intensity
        monitor.Error = np.sqrt(scale_old**2*monitor.Error**2+scale_new**2*new_monitor.Error**2)
        monitor.Ncount = monitor.Ncount + new_monitor.Ncount

        monitor.metadata.info["Ncount"] = total_ncount



def add_ncrystal_union_material(instrument, name, cfgstr, split=False):
    try:
        import NCrystal.mcstasutils as NCms
    except ImportError or ModuleNotFoundError:
        raise RuntimeError("Install NCrystal to use this feature in McStasScript.")

    abs_info, ncrystal_processes = NCms.cfgstr_2_unioncfg(cfgstr=cfgstr, split_by_physics=split)

    process_names = []
    for process in ncrystal_processes:
        process_name = name + "_" + process[1]
        process_names.append(process_name)
        comp = instrument.add_component(process_name, "NCrystal_process")
        comp.cfg = '"' + process[0] + '"'

    comp = instrument.add_component(name, "Union_make_material")
    comp.my_absorption = abs_info
    comp.process_string = '"' + ",".join(process_names) + '"'



from .simulation_interface import SimInterface
from .plot_interface import PlotInterface

from mcstasscript.interface.instr import McCode_instr
from mcstasscript.data.data import McStasData


def show(function_input):
    """
    Shortcut to showing an interface appropriate for given object

    Can show widget interface for instrument object or list of McStasData
    objects. Needs "%matplotlib widget" in notebook to work correctly.

    Parameters
    ----------

    function_input : McCode_instr, list of McStasData

    """

    if isinstance(function_input, McCode_instr):
        return show_instrunent(function_input)

    elif isinstance(function_input, (list, McStasData)):
        return show_plot(function_input)

    else:
        raise RuntimeError("Show did not recoignize object of type"
                           + str(type(function_input)) + ".")


def show_instrunent(instrument):
    """
    Shows simulation interface for instrument
    """
    simulation_interface = SimInterface(instrument)
    return simulation_interface.show_interface()


def show_plot(data):
    """
    Shows plot interface for given data
    """
    if isinstance(data, list):
        for element in data:
            if not isinstance(element, McStasData):
                raise RuntimeError("Given list contains elements that "
                                   + "are not McStasData objects.")
    else:
        if not isinstance(data, McStasData):
            raise RuntimeError("Given data is not McStasData.")

    plot_interface = PlotInterface(data)
    return plot_interface.show_interface()



import os
import io
import unittest.mock

from mcstasscript.helper.component_reader import ComponentReader


def setup_component_reader():
    """
    Sets up a component_reader instance with dummy mcstas
    installation located in the tests directory
    """

    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    component_reader = ComponentReader(mcstas_path=dummy_path)

    os.chdir(current_work_dir)  # Reset work directory

    return component_reader


def setup_component_reader_input_path():
    """
    Sets up a component_reader instance with dummy mcstas
    installation located in the tests directory, and specifies
    a input_path which is also located in the test directory.
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")
    input_path = os.path.join(THIS_DIR, "test_input_folder")

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    component_reader = ComponentReader(mcstas_path=dummy_path,
                                       input_path=input_path)

    os.chdir(current_work_dir)  # Reset work directory

    return component_reader


class TestComponentReader(unittest.TestCase):
    """
    Testing the ComponenReader class. As this class reads information
    from McStas, a dummy McStas install is made in the test folder to
    avoid the test results changing with updates of McStas.
    """

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_init_overwrite_message(self, mock_stdout):
        """
        Test that ComponentReader reports overwritten components
        Here using default input path, which is current folder
        """

        setup_component_reader()

        message = ("The following components are found in the work directory "
                   + "/ input_path:\n     test_for_reading.comp\n"
                   + "These definitions will be used instead of the "
                   + "installed versions.\n")

        self.assertEqual(mock_stdout.getvalue(), message)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_init_overwrite_message_input(self, mock_stdout):
        """
        Test that ComponentReader reports overwritten components
        Here using user defined input path, a directory in tests
        """

        setup_component_reader_input_path()

        message = ("The following components are found in the work directory "
                   + "/ input_path:\n     test_for_structure.comp\n"
                   + "These definitions will be used instead of the "
                   + "installed versions.\n")

        self.assertEqual(mock_stdout.getvalue(), message)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_init_filenames(self, mock_stdout):
        """
        Test that ComponentReader initializes component names correctly
        """

        component_reader = setup_component_reader()

        n_components_found = len(component_reader.component_path)
        self.assertEqual(n_components_found, 4)
        self.assertIn("test_for_reading", component_reader.component_path)
        self.assertIn("test_for_structure", component_reader.component_path)
        self.assertIn("test_for_structure2", component_reader.component_path)
        self.assertIn("test_for_skipped_sections", component_reader.component_path)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_init_component_paths(self, mock_stdout):
        """
        Test that ComponentReader stores correct absolute paths to
        the components found in the McStas installation
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")

        component_reader = setup_component_reader()

        n_components_found = len(component_reader.component_path)
        self.assertEqual(n_components_found, 4)

        expected_path = os.path.join(THIS_DIR, "test_for_reading.comp")
        path = component_reader.component_path["test_for_reading"]
        self.assertEqual(path, expected_path)

        category = component_reader.component_category["test_for_reading"]
        self.assertEqual(category, "work directory")

        expected_path = os.path.join(dummy_path, "misc",
                                     "test_for_structure.comp")
        self.assertIn("test_for_structure", component_reader.component_path)

        path = component_reader.component_path["test_for_structure"]
        self.assertEqual(path, expected_path)

        category = component_reader.component_category["test_for_structure"]
        self.assertEqual(category, "misc")

        self.assertIn("test_for_structure2", component_reader.component_path)

        path = component_reader.component_path["test_for_structure2"]
        expected_path = os.path.join(dummy_path, "sources",
                                     "test_for_structure2.comp")
        self.assertEqual(path, expected_path)

        category = component_reader.component_category["test_for_structure2"]
        self.assertEqual(category, "sources")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_init_component_paths_input(self, mock_stdout):
        """
        Test that ComponentReader stores correct absolute paths to
        the components found in the McStas installation.
        This version uses custom input_path
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")
        input_path = os.path.join(THIS_DIR, "test_input_folder")

        component_reader = setup_component_reader_input_path()

        n_components_found = len(component_reader.component_path)
        self.assertEqual(n_components_found, 4)

        self.assertIn("test_for_reading", component_reader.component_path)

        expected_path = os.path.join(dummy_path, "misc",
                                     "test_for_reading.comp")
        path = component_reader.component_path["test_for_reading"]
        self.assertEqual(path, expected_path)

        category = component_reader.component_category["test_for_reading"]
        self.assertEqual(category, "misc")

        expected_path = os.path.join(input_path, "test_for_structure.comp")
        self.assertIn("test_for_structure", component_reader.component_path)

        path = component_reader.component_path["test_for_structure"]
        self.assertEqual(path, expected_path)

        category = component_reader.component_category["test_for_structure"]
        self.assertEqual(category, "work directory")

        self.assertIn("test_for_structure2", component_reader.component_path)

        path = component_reader.component_path["test_for_structure2"]
        expected_path = os.path.join(dummy_path, "sources",
                                     "test_for_structure2.comp")
        self.assertEqual(path, expected_path)

        category = component_reader.component_category["test_for_structure2"]
        self.assertEqual(category, "sources")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_init_categories(self, mock_stdout):
        """
        Test that ComponentReader initializes categories correctly
        """

        component_reader = setup_component_reader()

        n_categories_found = len(set(component_reader.component_category.values()))
        self.assertEqual(n_categories_found, 3)
        """
        Categories are the values of the category dict, so n_categories 
        is the same as the set of the values.
        """
        category = component_reader.component_category["test_for_reading"]
        self.assertEqual(category, "work directory")
        category = component_reader.component_category["test_for_structure"]
        self.assertEqual(category, "misc")
        category = component_reader.component_category["test_for_structure2"]
        self.assertEqual(category, "sources")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_show_categories(self, mock_stdout):
        """
        This method prints to console, check it prints the categories
        in the dummy installation correctly.

        """
        component_reader = setup_component_reader()

        component_reader.show_categories()

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(len(output), 7)
        self.assertIn(" sources", output)
        self.assertIn(" work directory", output)
        self.assertIn(" misc", output)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_show_categories_ordered(self, mock_stdout):
        """
        Check that the print to console is ordered as usual. This test
        may be implementation dependent as python dictionaries are not
        ordered.

        """

        component_reader = setup_component_reader()

        component_reader.show_categories()

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(len(output), 7)
        # Ignoring message about overwritten components, starting from 3
        self.assertEqual(output[3], " misc")
        self.assertEqual(output[4], " sources")
        self.assertEqual(output[5], " work directory")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_show_components_short(self, mock_stdout):
        """
        Here we attempt to show components in the misc category.  In
        the dummy install, there are two components in this folder, but
        one of these is overwritten by the version in the current
        work directory.

        """

        component_reader = setup_component_reader()

        component_reader.show_components_in_category("misc", line_length=100)

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(len(output), 6)
        self.assertIn(" test_for_structure", output)
        # Check overwritten component is not in the output
        self.assertNotIn(" test_for_reading", output)

    """
    # This test not as important, but could be finished later
    @unittest.mock.patch("sys.stdout", new_callable = io.StringIO)
    def test_ComponentReader_show_components_long(self, mock_stdout):

        component_reader = setup_component_reader()

        #  Add elements directly to component_readers library
        # generate list
        # add list

        #component_reader.component_category[]

        component_reader.show_components_in_category("misc", line_length=100)

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(len(output), 3)
        self.assertIn(" test_for_structure", output)
    """

    # test load_all_components
    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_load_all_components(self, mock_stdout):
        """
        Load all components in the dummy install, but only one has any
        content.  The method is currently not necessary, as components
        are now loaded individually when needed.

        """

        component_reader = setup_component_reader()

        CompInfo_dict = component_reader.load_all_components()

        comp_name = "test_for_reading"
        name = CompInfo_dict[comp_name].name
        self.assertEqual(name, comp_name)

        parameter_names = CompInfo_dict[comp_name].parameter_names
        self.assertIn("target_index", parameter_names)
        parameter_types = CompInfo_dict[comp_name].parameter_types
        self.assertIn("target_index", parameter_types)
        parameter_defaults = CompInfo_dict[comp_name].parameter_defaults
        self.assertIn("target_index", parameter_defaults)

        type_str = CompInfo_dict[comp_name].parameter_types["target_index"]
        self.assertEqual(type_str, "int")

        comp_name = "test_for_structure"
        name = CompInfo_dict[comp_name].name
        self.assertEqual(name, comp_name)
        # test_for_structure is an empty file, so no conentet to check

        comp_name = "test_for_structure2"
        name = CompInfo_dict[comp_name].name
        self.assertEqual(name, comp_name)
        # test_for_structure2 is an empty file, so no conentet to check

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_read_name_error(self, mock_stdout):
        """
        read_name should throw an error when searching for a component
        that is not present in the installation.
        """

        component_reader = setup_component_reader()

        with self.assertRaises(NameError):
            component_reader.read_name("no_such_comp")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_read_name_success(self, mock_stdout):
        """
        Read component simply calls read_component_file, but here
        the output is checked against what is in the dummy file.

        """

        component_reader = setup_component_reader()

        CompInfo = component_reader.read_name("test_for_reading")

        self.assertEqual(CompInfo.name, "test_for_reading")
        self.assertEqual(CompInfo.category, "work directory")
        self.assertIn("dist", CompInfo.parameter_names)
        self.assertIn("dist", CompInfo.parameter_defaults)
        self.assertIn("dist", CompInfo.parameter_types)
        self.assertEqual(CompInfo.parameter_types["dist"], "double")
        self.assertIn("dist", CompInfo.parameter_comments)
        self.assertIn("dist", CompInfo.parameter_units)
        self.assertEqual(CompInfo.parameter_units["dist"], "m")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_find_components_names(self, mock_stdout):
        """
        Test that ComponentReader initializes component names correctly
        """

        component_reader = setup_component_reader()

        component_reader.component_path = {}
        component_reader.component_category = {}

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_path = os.path.join(THIS_DIR, "dummy_mcstas", "misc")
        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder
        component_reader._find_components(dummy_path)
        os.chdir(current_work_dir)  # Return to original work directory

        n_components_found = len(component_reader.component_path)
        self.assertEqual(n_components_found, 3)
        self.assertIn("test_for_reading", component_reader.component_path)
        self.assertIn("test_for_structure", component_reader.component_path)
        self.assertIn("test_for_skipped_sections", component_reader.component_path)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_find_components_categories(self, mock_stdout):
        """
        Test that ComponentReader initializes component categories correctly
        """

        component_reader = setup_component_reader()

        component_reader.component_path = {}
        component_reader.component_category = {}

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_path = os.path.join(THIS_DIR, "dummy_mcstas", "misc")
        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder
        component_reader._find_components(dummy_path)
        os.chdir(current_work_dir)  # Return to original work directory

        n_components_found = len(component_reader.component_category)
        self.assertEqual(n_components_found, 3)

        category = component_reader.component_category["test_for_reading"]
        self.assertEqual(category, "misc")
        category = component_reader.component_category["test_for_structure"]
        self.assertEqual(category, "misc")
        category = component_reader.component_category["test_for_skipped_sections"]
        self.assertEqual(category, "misc")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_read_component_category(self, mock_stdout):
        """
        Check that the correct category is returned.

        Can't run this test with overwritten component test_for_reading.
        read_component will report tests as category, but this is
        overwritten by read_name in normal use.
        """
        component_reader = setup_component_reader()

        path_for_test = component_reader.component_path["test_for_structure"]
        CompInfo = component_reader.read_component_file(path_for_test)

        exp_cat = component_reader.component_category["test_for_structure"]
        self.assertEqual(CompInfo.category, exp_cat)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_read_component_standard(self, mock_stdout):
        """
        Test that a normal parameter is read correctly when reading a
        component file.
        Has default, is double type, has comment, has unit
        """

        component_reader = setup_component_reader()

        path_for_test = component_reader.component_path["test_for_reading"]
        CompInfo = component_reader.read_component_file(path_for_test)

        self.assertIn("xwidth", CompInfo.parameter_names)

        self.assertIn("xwidth", CompInfo.parameter_defaults)
        self.assertEqual(CompInfo.parameter_defaults["xwidth"], 0.0)

        self.assertIn("xwidth", CompInfo.parameter_types)
        self.assertEqual(CompInfo.parameter_types["xwidth"], "double")

        self.assertIn("xwidth", CompInfo.parameter_comments)
        comment = "Width of rectangle test comment"
        self.assertEqual(CompInfo.parameter_comments["xwidth"], comment)

        self.assertIn("xwidth", CompInfo.parameter_units)
        self.assertEqual(CompInfo.parameter_units["xwidth"], "m")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_read_component_required(self, mock_stdout):
        """
        Test that a required parameter is read correctly when reading a
        component file.
        Has no default, is double type, has no comment, has no unit
        """

        component_reader = setup_component_reader()

        path_for_test = component_reader.component_path["test_for_reading"]
        CompInfo = component_reader.read_component_file(path_for_test)

        self.assertIn("gauss", CompInfo.parameter_names)

        self.assertIn("gauss", CompInfo.parameter_defaults)
        self.assertIsNone(CompInfo.parameter_defaults["gauss"])

        self.assertIn("gauss", CompInfo.parameter_types)
        self.assertEqual(CompInfo.parameter_types["gauss"], "double")

        self.assertNotIn("gauss", CompInfo.parameter_comments)

        self.assertNotIn("gauss", CompInfo.parameter_units)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_read_component_int(self, mock_stdout):
        """
        Test that an integer parameter is read correctly when reading a
        component file.
        Has default, is int type (comments and unit checked already)
        """

        component_reader = setup_component_reader()

        path_for_test = component_reader.component_path["test_for_reading"]
        CompInfo = component_reader.read_component_file(path_for_test)

        self.assertIn("flux", CompInfo.parameter_names)

        self.assertIn("flux", CompInfo.parameter_defaults)
        self.assertEqual(CompInfo.parameter_defaults["flux"], 1)

        self.assertIn("flux", CompInfo.parameter_types)
        self.assertEqual(CompInfo.parameter_types["flux"], "int")

        self.assertIn("flux", CompInfo.parameter_comments)
        # Have already tested comments are read

        self.assertIn("flux", CompInfo.parameter_units)
        # Have already tested units are read

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_read_component_string(self, mock_stdout):
        """
        Test that a string parameter is read correctly when reading a
        component file.
        Has no default, is string type (comments and unit checked already)
        """

        component_reader = setup_component_reader()

        path_for_test = component_reader.component_path["test_for_reading"]
        CompInfo = component_reader.read_component_file(path_for_test)

        self.assertIn("test_string", CompInfo.parameter_names)

        self.assertIn("test_string", CompInfo.parameter_defaults)
        self.assertIsNone(CompInfo.parameter_defaults["test_string"])

        self.assertIn("test_string", CompInfo.parameter_types)
        self.assertEqual(CompInfo.parameter_types["test_string"], "string")

        self.assertNotIn("test_string", CompInfo.parameter_comments)
        # Have already tested comments are read

        self.assertNotIn("test_string", CompInfo.parameter_units)
        # Have already tested units are read

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_ComponentReader_read_comp_with_skipped_sections(self, mock_stdout):
        """
        Read component simply calls read_component_file, but here
        the output is checked against what is in the dummy file.

        """

        component_reader = setup_component_reader()

        CompInfo = component_reader.read_name("test_for_skipped_sections")

        self.assertEqual(CompInfo.name, "test_for_skipped_sections")
        self.assertEqual(CompInfo.category, "misc")
        self.assertEqual(len(CompInfo.parameter_names), 4)
        self.assertIn("packing_factor", CompInfo.parameter_names)
        self.assertIn("packing_factor", CompInfo.parameter_defaults)
        self.assertIn("packing_factor", CompInfo.parameter_types)
        self.assertEqual(CompInfo.parameter_types["packing_factor"], "double")
        self.assertIn("packing_factor", CompInfo.parameter_comments)
        self.assertIn("packing_factor", CompInfo.parameter_units)
        self.assertEqual(CompInfo.parameter_units["packing_factor"], "1")


if __name__ == '__main__':
    unittest.main()



from mcstasscript.helper.mcstas_objects import DeclareVariable
from mcstasscript.helper.mcstas_objects import Component
from mcstasscript.instrument_diagnostics.diagnostics_instrument import DiagnosticsInstrument
from mcstasscript.instrument_diagnostics.view import View
from mcstasscript.instrument_diagnostics.plot_overview import PlotOverview
from mcstasscript.instrument_diagnostics import event_plotter
from mcstasscript.interface.functions import name_search

def sanitise_comp_name(comp_name):
    if isinstance(comp_name, Component):
        comp_name = Component.name

    if not isinstance(comp_name, str):
        raise ValueError("comp_name should be a string or component object!")

    return comp_name

class DiagnosticsPoint:
    def __init__(self, instr, comp_name, before=True, rays=50000):
        comp_name = sanitise_comp_name(comp_name)
        try:
            instr.get_component(comp_name)
        except KeyError:
            raise KeyError("Component '" + comp_name + "' not found in instrument.")

        self.component = comp_name
        self.before = before # False for after

        if isinstance(rays, str):
            if not rays == "all":
                raise ValueError("The ray keyword for a point should be all or an integer.")
        else:
            rays = int(rays)

        self.rays = rays

        # Attributes set by add_monitors
        self.filename = None
        # Attributes set by read_data
        self.recorded_rays = None

    def set_filename(self, filename):
        self.filename = filename

    def set_recorded_rays(self, rays):
        self.recorded_rays = rays

    def __eq__(self, other):
        return self.component == other.component and self.before == other.before

    def __repr__(self):
        string = "Diagnostics point "
        if self.before:
            string += "before: "
        else:
            string += "after:  "

        string += self.component.ljust(25)
        string += " - rays: "
        if self.recorded_rays is not None:
            string += str(self.recorded_rays)
            string += " / "
        string += str(self.rays)

        return string

class BeamDiagnostics(DiagnosticsInstrument):
    def __init__(self, instr):
        super().__init__(instr)

        # points to investigate with options
        self.points = []
        self.ordered_point_list = []

        # flags for monitor_nD user variables
        self.flags = []

        self.monitor_names = []

        self.views = []

        self.data = None

        self.event_plotters = []

    def __repr__(self):
        string = f"Instrument diagnostics for: {self.instr.name}\n"
        string += "Diagnostics points: \n"

        if len(self.points) == 0:
            string += "  No diagnostics points yet\n"
        for point in self.points:
            string += "  " + point.__repr__() + "\n"

        string += "Views: \n"
        if len(self.views) == 0:
            string += "  No views yet\n"

        for view in self.views:
            string += "  " + view.__repr__() + "\n"

        if len(self.flags) != 0:
            string += "Recording following user variables:\n"
        for flag in self.flags:
            string += "  " + str(flag) + "\n"

        if self.data is None:
            string += "Does not yet contain simulated data"
        else:
            string += "Does contain simulated data"

        return string

    def add_flag(self, flag):
        """
        Adds flag to be monitored in event list, maximum of 3 allowed

        The flag must be available in declared variables or user vars.

        Parameters:

        flag: str
            str with variable name to be recorded to event dataset
        """
        if not isinstance(flag, str):
            raise ValueError("flag has to be a string.")

        # find declare and user_var names in instrument
        names = [x.name for x in self.instr.declare_list
                 if isinstance(x, DeclareVariable)]
        names += [x.name for x in self.instr.user_var_list
                  if isinstance(x, DeclareVariable)]

        if flag not in names:
            raise ValueError(f"flag {flag} was not found in declared "
                             f"variables or uservars of the instrument.")

        if len(self.flags) >= 3:
            raise ValueError("Can't add more than three flags")

        self.flags.append(flag)

    def clear_flags(self):
        """
        Clears flags
        """
        self.flags = []

    def show_flags(self):
        """
        Shows list of current flags
        """
        if self.flags is not None:
            for index, flag in enumerate(self.flags):
                print(f" user{index+1}={flag}")

    def add_point(self, before=None, after=None, rays=50000):
        """
        Adds point in which the beam is investigated with given number of rays

        Can be before and/or after components, this is chosen by using the
        appropriate keyword argument.

        Parameters:

        before : str or component instance
            Reference to component beam should be monitored just before

        after : str or component instance
            Reference to component beam should be monitored just after

        rays : int or "all"
            Number of rays to record, either integer or the string "all"
        """
        if before is None and after is None:
            raise ValueError("Specify either before or after with component name.")

        if before is not None:
            self._add_diagnostics_point(comp_name=before, before=True, rays=rays)

        if after is not None:
            self._add_diagnostics_point(comp_name=after, before=False, rays=rays)

    def _add_diagnostics_point(self, comp_name, before, rays):
        """
        Internal function to add the DiagnosticsPoint to list
        """
        point = DiagnosticsPoint(self.instr, comp_name=comp_name, before=before, rays=rays)
        if point in self.points:
            # overwrite that point
            index = self.points.index(point)
            self.points[index] = point
        else:
            # add point to list
            self.points.append(point)

    def remove_point(self, before=None, after=None):
        """
        Removes a point from the list

        Specify whether the point in question is before or after the
        given component name or instance by using the appropriate
        keyword argument

        Parameter:

        before : str or component instance
            Name or component instance for which point before should be removed

        after : str or component instance
            Name or component instance for which point after should be removed
        """
        if before is None and after is None:
            raise ValueError("Specify either before or after with component name.")

        if before is not None:
            before = sanitise_comp_name(before)
            for index, point in enumerate(self.points):
                if point.component == before and point.before:
                    del self.points[index]

        if after is not None:
            after = sanitise_comp_name(after)
            for index, point in enumerate(self.points):
                if point.component == after and not point.before:
                    del self.points[index]

    def clear_points(self):
        """
        Remove all points
        """
        self.points = []

    def show_points(self):
        """
        Shows current diagnostics points
        """
        for point in self.points:
            print(point)

    def add_view(self, axis1, axis2=None, bins=100, same_scale=False, **kwargs):
        """
        Add a view with one or two axes that will be shown at all diagnostics points

        Parameters:
        axis1: str
            Name of parameter for first axis

        axis2: str
            Name of parameter for second axis

        bins : int or list of length 2
            Number of bins for histogram (can be list of length 2 for 2D)

        same_scale : bool
            Select if the same scale should be ensured for all points with this view
        """
        view = View(axis1=axis1, axis2=axis2, bins=bins, same_scale=same_scale, **kwargs)
        self.views.append(view)

    def clear_views(self):
        """
        Clear list of views
        """
        self.views = []

    def show_views(self):
        """
        Shows views
        """
        for view in self.views:
            print(view)

    def add_monitors(self):
        """
        Adds monitors as specified by list of diagnostics points

        The method sorts the points with the component sequence so the
        resulting plots are in the correct order.
        """
        comp_names = [x.name for x in self.instr.make_component_subset()]
        self.monitor_names = []
        self.ordered_point_list = []
        point_names = [x.component for x in self.points]

        for name in comp_names:
            before_point = None
            after_point = None
            if name in point_names:
                for point in self.points:
                    if point.component == name:
                        if point.before:
                            before_point = point
                        else:
                            after_point = point

            if before_point is not None:
                self.ordered_point_list.append(before_point)
            if after_point is not None:
                self.ordered_point_list.append(after_point)

        for point in self.ordered_point_list:
            comp = self.instr.get_component(point.component)
            filename = self.add_monitor(point, comp)
            point.set_filename(filename)

    def add_monitor(self, point, comp_instance):
        """
        Adds monitor_nD event monitor to diagnostics instrument for given point

        Needs point and component instance that match

        Parameters:

        point : DiagnosticsPoint
            Point for which monitor should be added

        comp_instance : Component
            Component instance relative to which the monitor should be placed
        """

        if not point.component == comp_instance.name:
            raise RuntimeError("Given point and component instance should match!")

        user_vars = [None, None, None]
        flags_str = ""
        for index, flag in enumerate(self.flags):
            flags_str += " user" + str(index+1)
            if self.instr.mccode_version == 2:
                # Monitor_nD in McStas 2.X needs to be a raw variable
                user_vars[index] = flag
            else:
                # Monitor_nD in McStas 3.X (and onwards) needs to be a string
                user_vars[index] = '"' + flag + '"'

        if isinstance(point.rays, (float, int)):
            ray_value = point.rays
            if self.instr.mccode_version == 3:
                ray_value += 1 # In McStas 3.0 Monitor_nD grabs one event too few
            ray_str = str(int(ray_value)) # Monitor_nD seems to record one ray less than requested
        else:
            ray_str = point.rays # Case of rays set to all

        if point.before:
            name = "Diag_before_" + point.component
            options = f'"square boarders n x y z vx vy vz t{flags_str}, list {ray_str}"'
            mon = self.instr.add_component(name, "Monitor_nD", before=point.component)
        else:
            name = "Diag_after_" + point.component
            options = f'"previous n x y z vx vy vz t{flags_str}, list {ray_str}"'
            mon = self.instr.add_component(name, "Monitor_nD", after=point.component)

        mon.set_parameters(restore_neutron=1,
                           xwidth=100, yheight=100,
                           options=options,
                           user1=user_vars[0], user2=user_vars[1], user3=user_vars[2],
                           filename='"' + name + ".diag" + '"')


        mon.set_AT(comp_instance.AT_data, RELATIVE=comp_instance.AT_reference)
        if comp_instance.ROTATED_specified:
            mon.set_ROTATED(comp_instance.ROTATED_data, RELATIVE=comp_instance.ROTATED_reference)

        return name

    def run(self):
        """
        Runs diagnostics with all included points

        Saves data in data attribute and the resulting instrument can be found
        in the instr attribute
        """
        self.reset_instr()
        self.remove_previous_use()

        self.add_monitors()

        self.correct_target_index()

        self.data = self.instr.backengine()

        self.read_data()

    def read_data(self):
        """
        Reads the generated data and organizes it as event plotter instances
        """

        self.event_plotters = []

        for point in self.ordered_point_list:
            try:
                event_data = name_search(point.filename, self.data)
            except NameError:
                # Data was not generated, no neutrons reached it or simulation failed
                continue

            point.set_recorded_rays(event_data.metadata.total_N)
            plotter = event_plotter.EventPlotter(point.filename, event_data,
                                                 flag_info=self.flags)

            self.event_plotters.append(plotter)

    def plot(self):
        """
        Plots the generated data for all points with all views
        """

        if len(self.event_plotters) == 0:
            if len(self.views) == 0:
                print("No data to plot! Add views and run.")
            else:
                print("No data to plot! Use the run method to generate data.")

        overview = PlotOverview(self.event_plotters, self.views)
        overview.plot_all()




import os


def work_dir_test(func):
    def wrapper():
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        current_work_dir = os.getcwd()

        os.chdir(THIS_DIR)  # Set work directory to test folder

        func()

        os.chdir(current_work_dir)  # Return to previous workdir

    return wrapper



# Demonstration of McStasScript, an API for creating and running McStas instruments from python scripts
# Written by Mads Bertelsen, ESS DMSC
import random

import mcstasscript as ms

# Create a McStas instrument
instr = ms.McStas_instr("random_demo",
                        author = "Mads Bertelsen",
                        origin = "ESS DMSC")

# Set up a material called Cu with approrpiate properties
# (uses McStas Union components, here the processes)

Init = instr.add_component("init", "Union_init")

Cu_inc = instr.add_component("Cu_incoherent", "Incoherent_process")
Cu_inc.sigma = 4*0.55
Cu_inc.packing_factor = 1
Cu_inc.unit_cell_volume = 55.4

Cu_powder = instr.add_component("Cu_powder", "Powder_process")
Cu_powder.reflections = "\"Cu.laz\""

Cu = instr.add_component("Cu", "Union_make_material")
Cu.my_absorption = "100*4*3.78/55.4"
Cu.process_string = "\"Cu_incoherent,Cu_powder\""

# Add neutron source
Source = instr.add_component("source", "Source_div")
# Add parameter to select energy at run time
energy = instr.add_parameter("double", "energy", value=10, comment="[meV] source energy")

Source.xwidth = 0.12
Source.yheight = 0.12
Source.focus_aw = 0.1
Source.focus_ah = 0.1
Source.E0 = energy
Source.dE = 0.0
Source.flux = 1E13

# List of available materials, Vacuum is provided by the system
material_name_list = ["Cu", "Vacuum"]

# Wish to set up a number of random boxes, here the number is chosen at random
number_of_volumes = random.randint(30, 40)

# Initialize the priority that needs to be unique for each volume
current_priority = 99
for volume in range(number_of_volumes):

    current_priority = current_priority + 1 # update the priority
    max_side_length = 0.04
    max_depth = 0.003
    # Set position in 10x10x10 cm^3 box 1 m from source
    position = [random.uniform(-0.05,0.05),
                random.uniform(-0.05,0.05),
                1+random.uniform(-0.05,0.05)]
    # Set random rotation
    rotation = [random.uniform(0,360),
                random.uniform(0,360),
                random.uniform(0,360)]

    # Choose a random material from the list of available materials
    volume_material = random.choice(material_name_list)

    # Add a McStas Union geometry with unique name
    this_geometry = instr.add_component("volume_" + str(volume), "Union_box")
    this_geometry.xwidth = random.uniform(0.01, max_side_length)
    this_geometry.yheight = random.uniform(0.01, max_side_length)
    this_geometry.zdepth = random.uniform(0.01, max_side_length)
    this_geometry.material_string = '"' + volume_material + '"'
    this_geometry.priority = current_priority
    this_geometry.p_interact = 0.3
    
    this_geometry.set_AT(position, RELATIVE="ABSOLUTE")
    this_geometry.set_ROTATED(rotation, RELATIVE="ABSOLUTE")

# A few Union loggers are set up for display of the scattering locations
space_2D_zx = instr.add_component("logger_space_zx_all", "Union_logger_2D_space")
space_2D_zx.filename = '"space_zx.dat"'
space_2D_zx.n1 = 1000
space_2D_zx.D_direction_1 = '"z"'
space_2D_zx.D1_min = -0.05
space_2D_zx.D1_max = 0.05
space_2D_zx.n2 = 1000
space_2D_zx.D_direction_2 = '"x"'
space_2D_zx.D2_min = -0.05
space_2D_zx.D2_max = 0.05
space_2D_zx.set_AT([0,0,1])

space_2D_zy = instr.add_component("logger_space_zy_all", "Union_logger_2D_space")
space_2D_zy.filename = '"space_zy.dat"'
space_2D_zy.n1 = 1000
space_2D_zy.D_direction_1 = '"z"'
space_2D_zy.D1_min = -0.05
space_2D_zy.D1_max = 0.05
space_2D_zy.n2 = 1000
space_2D_zy.D_direction_2 = '"y"'
space_2D_zy.D2_min = -0.05
space_2D_zy.D2_max = 0.05
space_2D_zy.set_AT([0,0,1])

# Union master component that executes the simulation of the random boxes
instr.add_component("random_boxes", "Union_master")

# McStas monitors for viewing the beam after the random boxes
PSD = instr.add_component("detector", "PSD_monitor", AT=[0,0,2])
PSD.xwidth = 0.1
PSD.yheight = 0.1
PSD.nx = 500
PSD.ny = 500
PSD.filename = '"PSD.dat"'
PSD.restore_neutron = 1

big_PSD = instr.add_component("large_detector", "PSD_monitor", AT=[0,0,2])
big_PSD.xwidth = 1.0
big_PSD.yheight = 1.0
big_PSD.nx = 500
big_PSD.ny = 500
big_PSD.filename = '"big_PSD.dat"'
big_PSD.restore_neutron = 1

Stop = instr.add_component("stop", "Union_stop")

instr.show_components()

# Run the McStas simulation, a unique foldername is required for each run
instr.settings(output_path="demonstration", mpi=2, ncount=5E7)
instr.set_parameters(energy=600)

data = instr.backengine()

# Set plotting options for the data (optional)
ms.name_plot_options("logger_space_zx_all", data, log=1, orders_of_mag=3)
ms.name_plot_options("logger_space_zy_all", data, log=1, orders_of_mag=3)
ms.name_plot_options("detector", data, log=1, colormap="hot", orders_of_mag=0.5)
ms.name_plot_options("large_detector", data, log=1, orders_of_mag=8)

# Plot the resulting data on a logarithmic scale
plot = ms.make_sub_plot(data)
